# -*- coding: utf-8 -*-
from .STILParser import STILParser

from .WFCUtils import WFCUtils
from .DomainUtils import DomainUtils
from .PattVecCmd import PattVecCmd
from .HashInfo import HashInfo
from .STILDumpCompilerExceptions import STILDumpCompilerException

import os

'''
This is an example class which can be used as a base for a real STIL compiler.
The main compile method will make a syntax and semantic analisys of the file
and will dump the content of the STIL file in several text files:
    
pattern_blocks.flow :   this is the main entry point which define the order of 
                        pattern blocks execution (in *.pattern_blocks files), 
                        depends on information found in PatternExec/PatternBurst 
                        blocks.

*.pattern_blocks    :   one or more files with "compiled" information from the 
                        pattern blocks. The information includes : one or more 
                        lines for every vector statement as vector address with 
                        information for WaveformCharacter for every used signal 
                        and pattern statement as separate command.
                        The format is the following:
                        #                                               -> start of comment
                        SIGNALS_ORDER|sig1+...+sigN                     -> order of the signal used to decode WFC to signal information
                        VA|VAP|VAMP|WFCS|CMD|CMD_ARG| ... |CMD|CMD_ARG  -> one vector address
                        Where:
                            |    is a data separator.
                            VA   is the absolute vector address for the particular 
                                 pattern block.
                            VAP  is the absolute vector address for every pattern 
                                 statement in the pattern block if there is no 
                                 Macro or expanding Procedure invocation in this 
                                 pattern block. 
                                 Value '-' indicates that there is WFC/CMD data 
                                 from Macro/Procdure at this VA
                            VAMP is absolute vector address of the Macro or expanded Procedure block.
                                 Value '-' indicates that there is no WFC/CMD 
                                 data from Macro/Procdure at this VA
                            WFCS is a ordered WaveformCharacter for every used 
                                 signal in the pattern block. The order of signals is defined in SIGNALS_ORDER line
                            CMD  Short name of the command. The command is based 
                                 on found pattern statement in the pattern block. 
                                 One pattern statement can generate one or more commands.
                                 For example : 
                                     "Loop X { ... }" statement will generate three commands:
                                         "LLC" (Load Loop Counter) with value X 
                                                 which indicates how many times 
                                                 the block will repeat.
                                         "SP" (Start Loop) on the first "Vector" 
                                                 statement after the "Loop" statement. 
                                                 This command defines the address where 
                                                 the EL must jump if the X counter 
                                                 value not reached.
                                         "EL" (End Loop) which will be placed on 
                                                 the last vector of the Loop block 
                                                 and will jump to SP address
                                     "Stop" statement will generate just one command - "E" (stands for End)
                            CMD_ARG : value of the CMD if applicable. For example the counter for the "LLC" command (Load Loop Counter)
                            Note : for one vector address, multiple commands can exists. 
                            They will be separated with "|".
                            The full description of the commands is available in every generated pattern_blocks file.
                            
*.procedure_block   :   one or more files with "compiled" information from the 
                        procedures blocks if the option "expanding_procs" is 
                        set to False.
                        The format is the same as in the *.pattern_blocks
                        
pattern_name.sm     :   Scan memory file which contains WFC sequence for every 
                        signal used in Procedure or Macro invocation. Depends on
                        locations and number of # in the procedure or macro bodies,
                        not all WFC in the procdure/macro call will be available
                        in the .sm file due WFC substition and normalization
                        according the STIL standard.

timing.txt          :   Text file with timing information 
                        
Class's arguments (passed as class constructor arguments):
    
    stil_file               : location of the input stil file. Mandatory
    
    expanding_procs         : Expanding or not the vectors from procedures into
                              the pattern block where the procedure is called.
                              Expanding is useful if the ATE pattern engine 
                              lack of capabilities to execute real procedures. 
                              This option impacts calculation of the vector addreses.
                              
    is_scan_mem_available   : depends on ATE pattern engine capabilities,
                              WFC used as arguments in the Macro/Proc invocation
                              can be stored into a dedicated scan memory or to 
                              be expanded in the general available memory.
                              If the option is False, all WFC will be expanded
                              into vector statements in the pattern block where
                              procedures or macros with "Shift" statements are
                              invoked.
                              This option impacts calculation of the vector addreses.
    
    out_folder              : output folder where files will be generated. If
                              not provided, they will be generated in the current
                              working folder
    
    signals_order_file      : file with custom order of the signals. One line - one signal name
    
Restrictions:
    Because ATE pattern engine can have many restrictions (in the HW or in the FW),
    our dump compiler is based on imaginary HW/FW with the following capabilities 
    and restrictoins:
        1. One vector address can hold unlimited commands.
        2. Unlimited number of:
            - Signals
            - Named Signals groups
            - Named Timing blocks
            - Waveform tables in one Timing block
            - Waveform characters (WFC) per signals 
            - Waveform events (WFE) per WFC at any place within one period.
              All defined in IEEE 1450.0 WFE are supported.
            - Vector addresses
            - Procedures calls
            - Named PatternExec/PatternBurts blocks
            - Named Pattern blocks
        3. No restrictions for the period value in any Waveform table.
        4. Call to a procedure command is always placed in the vector preceded
           the "Call" statement. This means "Call" statement, before any "Vector" 
           statement will throw a compilation error. This restriction is not
           valid if option "expanding_procs" is set to value "True".
           The reason behind this restriction is to use the prevoiuos vector 
           statement as place to put the FW command for preparation the jump to
           the procedure's first address.
        5. Call to a procedure statement is not allowed immediately after end of 
           Loop/MatchLoop statement block. The reason is again the need for a 
           vector to place the "jump to procedure" command. If such command is 
           placed in the last vector of the Loop/MatchLoop command, it will 
           always jump to the procedure at the last vector of the Loop/MatchLoop 
           block instead the first vector after the Loop/MatchLoop block.
        6. Nested Loop/MatchLoop blocks are not allowed. Compilation error will be thrown.
        7. Loop/MatchLoop statements need a command to set the number of loops.
           This must be done before the first vector where the Loop/MatchLoop 
           block statements start. Because of this, the following cases, will 
           thrown compilation errors:
               - There is no Vector statement before end of one Loop/MatchLoop 
                 block and the next one.
               - When the start of the Loop/MatchLoop block is the first statement
                 before any "Vector" statement.
        8. Backward GoTo statements (go to the label, prevoiusly defined) will 
           trown a compilation error. This is to prevent an endless cycle.
        9. In the case when "#" (incremental data substitution) is before or 
           after the Shift block and the option "expanding_procs" is False, the 
           compiler will thrown an error, because we assume that scan memory can
           be used only in the shift block. Otherwise every call to a not expanded 
           procedure have to set WFC values in the pre- and post- shift occurance 
           of the "#".

'''

class STILDumpCompiler(STILParser):
    def __init__(
        self,
        stil_file,
        out_folder,
        signals_order_file = None,
        propagate_positions = True,
        expanding_procs = False,
        is_scan_mem_available = False,
        dump_data = True,
        enable_trace = True,
        stil_lark_file = None,
        extra_grammars = [],
        debug = False,
    ):

        STILParser.__init__(self, stil_file, propagate_positions, expanding_procs, stil_lark_file, extra_grammars, debug)
        # Depends on ATE pattern engine capabilities, WFC used as arguments in the Macro/Proc invocation
        # can be stored into a dedicated scan memory or to be expanded in the general available memory
        # This choice impacts calculation of the total VA
        self.is_scan_mem_available = is_scan_mem_available
        
        # In the case when the STILDumpCompiler is used as base for a real compiler
        # this option will supprass the creation of the text files by STILDumpCompiler
        self.dump_data = dump_data
        
        self.enable_trace = enable_trace
        
        out_folder = os.path.join(os.getcwd(), out_folder)
        try: 
            os.mkdir(out_folder)
        except OSError: 
            pass

        self.out_folder = out_folder
        
        self.signals_order = []
        
        if signals_order_file != None:
            if os.path.exists(signals_order_file):
                with open(signals_order_file) as f:
                    self.signals_order.append(f.readline())
            else:
                raise Exception("File with signal order does not exists")
        # Key is in format SignalsGroupDomain::Pattern
        # Value is a dict with :
        #    key - signal name
        #    value is a list with WFC for this signal
        self.sgd_patt2sig2WFCs = {}

        # Number of wfc for the first vector address for each pattern block
        #    key - pattern block name
        #    value - number of wfcs
        self.patt2wfcs_number = {}

        # Key is in format MacroDefsDomain::Macro
        # Value is a list with all SignalsGroup domains used by macro
        self.macro2signal_group_domains = {}
        # Key is in format SignalsGroupDomain::MacroDefsDomain::Macro
        # Value is a dict with :
        #    key - signal name
        #    value is a list with WFC for this signal
        self.sgd_macro2sig2WFCs = {}
        # Number of wfc for the first vector address for each MacroDefs::Macro block
        #    key - MacroDefs block domain :: Macro block name
        #    value - number of wfcs
        self.macro2wfcs_number = {}

        # Key is in format ProceduresDomain::Procedure
        # Value is a list with all SignalsGroup domains used by procedure
        self.proc2signal_group_domains = {}
        # Key is in format SignalsGroupDomain::ProceduresDomain::Procedure
        # Value is a dict with :
        #    key - signal name
        #    value is a list with WFC for this signal
        self.sgd_proc2sig2WFCs = {}
        # Number of wfc for the first vector address for each Procedures::Procedure block
        #    key - Procedures block domain :: Procedure block name
        #    value - number of wfcs
        self.proc2wfcs_number = {}


        # dict key   is the pattern name
        # dict value is a list of command lists like:
        # pattern_name : [[VECTOR_0, LABEL, WFT], [VECTOR_1], ..., [CALL]]
        # In one vector address we can have multiple commands
        self.patt2cmd = {}

        # dict key   is the macro name in full domain format -> domain::macro_name
        # dict value is a list of command lists like:
        # pattern_name : [[VECTOR_0, LABEL, WFT], [VECTOR_1], ..., [CALL]]
        # In one vector address we can have multiple commands.
        self.macro2cmd = {}

        # dict key   is the procedure name in full domain format -> domain::procedure_name
        # dict value is a list of command lists like:
        # pattern_name : [[VECTOR_0, LABEL, WFT], [VECTOR_1], ..., [CALL]]
        # In one vector address we can have multiple commands and 
        # in the case of not expanded Procedures we do not have 
        # vector address at all, because the ATE engine must invoke 
        # the procedure code from separate addresss space
        self.proc2cmd = {}

        # Key is in format SignalsGroupDomain::ProceduresDomain::Procedure
        # Value is a dict with :
        #    key - signal name in the condition statamenet
        #    value is a list with WFC for this signal
        #self.sgd_proc2cond_sig2WFCs = {}

        # Current pattern command : PattVecCmd object
        self.curr_cmd = None
        
#        self.is_shift = False
#        self.is_condition_stmt = False
        self.is_fixed_stmt = False

        # Temp dict with :
        # Key is in format SignalsGroupDomain::PatternName
        # Value is a list with all signals for this pair of key
        self.sgd_patt2signals = {}
        # Key is in format SignalsGroupDomain::PatternName::SignalName
        # Value is the default WFC for this signal, taken either from Condition 
        # statment or from the first Vector statement
        self.sgd_patt_sig2defWFC = {}
        
        # Last signal to WFC info, common for pattern/macro/proc
        # key is signals group domain 
        # value is a dict with key signal name and value WFC 
        self.sgd2last_sig2wfc = {}

        # WFC before the first # or % per signals, common for pattern/macro/proc
        # Used when WFC of a signal is not defined after # or %. 
        # key is signals group domain 
        # value is a dict with key signal name and value WFC 
        self.sgd2before_subs_sig2wfc = {}

        # dict key   is the procedure name in full domain format -> domain::procedure_name
        # dict value is the number of vector addresses used by the procedure
        self.proc2vas = {}

        # scan data is a list with maps which contains :
        # key - timing_domain::wft_name::signal name
        # value - WFC list 
        self.scan_data = []

        # data contains list of lists with :
        # [0]     VA - list with vector addresses
        # [1]     | separator 
        # [2]     VAP - list with pattern only vector addresses
        # [3]     | separator 
        # [4]     VAMP -list with macro/proc only vector addresses
        # [5]     | separator 
        # [6+X-1]   For every used signal (total X signals), separate list with own WFCs
        # [6+X]   | separator 
        # [6+X+1] one or more commands for every vector address, separated with |. Temporary, it will be replaced by next record:
        # [6+X+2] PattVecCmd for every vector. Late
        self.data = []
        
        self.is_matchloop_inf = False

    def compile(self):

        if self.enable_trace:
            print("\nSyntax parsing is in progress...")
        self.parse_syntax()
        if self.enable_trace:
            print("Semantic analisys is in progress...")
        self.parse_semantic()
        # The real compilation will happen once the semantic
        # parsing is finished in self.eof() method bellow

    def eof(self):
        # at the end of the file:
        if self.enable_trace:
            print("Dump compilation is in progress...")
        try:
            self.dump_timing_data()
            self.dump_pattern_flow()
            if self.expanding_procs == False:
                 self.dump_procedures()
            patt_va, proc_va = self.dump_pattern_blocks()
            self.assamble_all()

            if self.enable_trace:
                print(f"Vector statements in pattern blocks : {patt_va}")
                if self.expanding_procs:
                    print(f"Vector statements in procedures blocks : {proc_va} (Procedure memory option is disabled) ")
                else:
                    print(f"Vector statements in procedures blocks : {proc_va}")
                print("Compiled with the following options:")
                if self.is_scan_mem_available:
                    print(" => Scan memory option is enabled [is_scan_mem_available=True] ")
                else:
                    print(" => Scan memory option is disabled (shift operator is expanded) [is_scan_mem_available=False] ")
                if self.expanding_procs:
                    print(" => Procedure memory option is not available [expanding_procs=True]")
                else:
                    print(" => Procedure memory option is available [expanding_procs=False]")
    
                print("Compilation is finished.")

        except STILDumpCompilerException as e:
            print(e.msg)
            os._exit(1)
            
    def add_cmd_patt(self, cmd, value=None):

        if self.curr_cmd == None:
            self.curr_cmd = PattVecCmd()
        self.curr_cmd.add_cmd(cmd, value)

    def add_prop_cmd_patt(self, cmd, prop, value):
        if self.curr_cmd == None:
            self.curr_cmd = PattVecCmd()
        self.curr_cmd.add_prop(cmd, prop, value)

    def save_cmd_patt(self):
        patt_cmd = self.patt2cmd[self.curr_pattern]
        patt_cmd.append(self.curr_cmd)
        self.curr_cmd = None

    def save_in_prev_cmd_patt(self, cmd, value = None):
        patt_cmd = self.patt2cmd[self.curr_pattern]
        vec_cmd = patt_cmd[-1]
        vec_cmd.add_cmd(cmd, value)

    def add_cmd_macro(self, cmd, value=None):
        if self.curr_cmd == None:
            self.curr_cmd = PattVecCmd()
        self.curr_cmd.add_cmd(cmd, value)        

    def save_cmd_macro(self):
        patt_cmd = self.macro2cmd[self.curr_macro_name]
        patt_cmd.append(self.curr_cmd)
        self.curr_cmd = None

    def save_in_prev_cmd_macro(self, cmd, value = None):
        patt_cmd = self.macro2cmd[self.curr_macro_name]
        vec_cmd = patt_cmd[-1]
        vec_cmd.add_cmd(cmd, value)

    def add_cmd_proc(self, cmd, value=None):
        if self.curr_cmd == None:
            self.curr_cmd = PattVecCmd()
        self.curr_cmd.add_cmd(cmd, value)
        
    def save_cmd_proc(self):
        patt_cmd = self.proc2cmd[self.curr_proc_name]
        patt_cmd.append(self.curr_cmd)
        self.curr_cmd = None

    def save_in_prev_cmd_proc(self, cmd, value = None):
        patt_cmd = self.proc2cmd[self.curr_proc_name]
        vec_cmd = patt_cmd[-1]
        vec_cmd.add_cmd(cmd, value)

    def b_signals__CLOSE_SIGNAL_BLOCK(self, t):
        if len(self.signals_order) > 0:
            for existing_signal in self.sig2type:
                if existing_signal not in self.signals_order:
                    raise Exception(f"Can not follow the signal order in {self.signals_order_file} due missing signal in signal block")
    
    def b_pattern__pattern_statements__WAVEFORM_TABLE_NAME(self, t):
        super().b_pattern__pattern_statements__WAVEFORM_TABLE_NAME(t)
        self.add_cmd_patt(PattVecCmd.CMD_WFT, self.curr_wft)

    def b_macrodefs__pattern_statements__WAVEFORM_TABLE_NAME(self, t):
        super().b_macrodefs__pattern_statements__WAVEFORM_TABLE_NAME(t)
        self.add_cmd_macro(PattVecCmd.CMD_WFT, self.curr_wft)

    def b_procedures__pattern_statements__WAVEFORM_TABLE_NAME(self, t):
        super().b_procedures__pattern_statements__WAVEFORM_TABLE_NAME(t)
        self.add_cmd_proc(PattVecCmd.CMD_WFT, self.curr_wft)

    def b_pattern__pattern_statements__CALL_VEC_DATA_STRING(self, t):
        ewfc = WFCUtils.expand_wfcs(t.value)
        self.add_prop_cmd_patt(PattVecCmd.CMD_CALL, self.curr_sig_ref, ewfc)
        super().b_pattern__pattern_statements__CALL_VEC_DATA_STRING(t)

    def b_pattern__pattern_statements__MACRO_VEC_DATA_STRING(self, t):
        ewfc = WFCUtils.expand_wfcs(t.value)
        self.add_prop_cmd_patt(PattVecCmd.CMD_MACRO, self.curr_sig_ref, ewfc)
        super().b_pattern__pattern_statements__MACRO_VEC_DATA_STRING(t)

    def b_pattern__pattern_statements__KEYWORD_C(self, t):
        self.is_condition_stmt = True

    def b_pattern__pattern_statements__KEYWORD_CONDITION(self, t):
        self.is_condition_stmt = True

    def b_pattern__pattern_statements__KEYWORD_F(self, t):
        self.is_fixed_stmt = True

    def b_pattern__pattern_statements__KEYWORD_FIXED(self, t):
        self.is_fixed_stmt = True

    def b_pattern__pattern_statements__KEYWORD_V(self, t):
        super().b_pattern__pattern_statements__KEYWORD_V(t)
        self.add_cmd_patt(PattVecCmd.CMD_VECTOR)
        self.is_vector_stmt = True

    def b_macrodefs__pattern_statements__KEYWORD_V(self, t):
        super().b_macrodefs__pattern_statements__KEYWORD_V(t)
        self.add_cmd_macro(PattVecCmd.CMD_VECTOR)

    def b_procedures__pattern_statements__KEYWORD_V(self, t):
        super().b_procedures__pattern_statements__KEYWORD_V(t)
        self.add_cmd_proc(PattVecCmd.CMD_VECTOR)

    def b_pattern__pattern_statements__KEYWORD_VECTOR(self, t):
        super().b_pattern__pattern_statements__KEYWORD_VECTOR(t)
        self.add_cmd_patt(PattVecCmd.CMD_VECTOR)
        self.is_vector_stmt = True

    def b_macrodefs__pattern_statements__KEYWORD_VECTOR(self, t):
        super().b_macrodefs__pattern_statements__KEYWORD_VECTOR(t)
        self.add_cmd_macro(PattVecCmd.CMD_VECTOR)

    def b_procedures__pattern_statements__KEYWORD_VECTOR(self, t):
        super().b_procedures__pattern_statements__KEYWORD_VECTOR(t)
        self.add_cmd_proc(PattVecCmd.CMD_VECTOR)

    def b_pattern__pattern_statements__LABEL(self, t):
        label = super().b_pattern__pattern_statements__LABEL(t)
        self.add_cmd_patt(PattVecCmd.CMD_LABEL, label)

    def b_macrodefs__pattern_statements__LABEL(self, t):
        label = super().b_macrodefs__pattern_statements__LABEL(t)
        self.add_cmd_macro(PattVecCmd.CMD_LABEL, label)
        
    def b_procedures__pattern_statements__LABEL(self, t):
        label = super().b_procedures__pattern_statements__LABEL(t)
        self.add_cmd_proc(PattVecCmd.CMD_LABEL, label)

    def b_pattern__pattern_statements__GOTO_LABEL(self, t):
        super().b_pattern__pattern_statements__GOTO_LABEL(t)
        self.save_in_prev_cmd_patt(PattVecCmd.CMD_GOTO, t.value)

    def b_macrodefs__pattern_statements__GOTO_LABEL(self, t):
        super().b_macrodefs__pattern_statements__GOTO_LABEL(t)
        self.save_in_prev_cmd_macro(PattVecCmd.CMD_GOTO, t.value)

    def b_procedures__pattern_statements__GOTO_LABEL(self, t):
        super().b_procedures__pattern_statements__GOTO_LABEL(t)
        self.save_in_prev_cmd_proc(PattVecCmd.CMD_GOTO, t.value)

    def b_macrodefs__pattern_statements__VEC_DATA_STRING(self, t):
        super().b_macrodefs__pattern_statements__VEC_DATA_STRING(t)

    def b_procedures__pattern_statements__VEC_DATA_STRING(self, t):
        super().b_procedures__pattern_statements__VEC_DATA_STRING(t)
        
    # to be moved to WFC decoding where # is !!!
    def b_macrodefs__pattern_statements__KEYWORD_SHIFT(self, t):
        super().b_macrodefs__pattern_statements__KEYWORD_SHIFT(t)
        self.add_cmd_macro(PattVecCmd.CMD_START_SHIFT)

    def b_procedures__pattern_statements__KEYWORD_SHIFT(self, t):
        super().b_procedures__pattern_statements__KEYWORD_SHIFT(t)
        self.add_cmd_proc(PattVecCmd.CMD_START_SHIFT)

    def b_macrodefs__pattern_statements__close_shift_block(self, t):
        super().b_macrodefs__pattern_statements__close_shift_block(t)
        self.save_in_prev_cmd_macro(PattVecCmd.CMD_STOP_SHIFT)

    def b_procedures__pattern_statements__close_shift_block(self, t):
        super().b_procedures__pattern_statements__close_shift_block(t)
        self.save_in_prev_cmd_proc(PattVecCmd.CMD_STOP_SHIFT)

    def b_pattern__pattern_statements__CALL_PROC_NAME(self, t):
        super().b_pattern__pattern_statements__CALL_PROC_NAME(t)
        patt_cmd = self.patt2cmd[self.curr_pattern]
        #is_empty = (len(patt_cmd) == 0)
        is_not_vec = (patt_cmd[-1].have_cmd(PattVecCmd.CMD_VECTOR) == False)
        is_not_macro = (patt_cmd[-1].have_cmd(PattVecCmd.CMD_MACRO) == False)
        if self.expanding_procs == False and is_not_vec and is_not_macro:
            err_msg = "COMPILER ERROR: At least one Vector statement is required before Call statement!"
            raise Exception(err_msg)
        else:
            self.add_cmd_patt(PattVecCmd.CMD_CALL, self.curr_proc_name_call)

    def b_macrodefs__pattern_statements__CALL_PROC_NAME(self, t):
        super().b_macrodefs__pattern_statements__CALL_PROC_NAME(t)
        patt_cmd = self.macro2cmd[self.curr_macro_name]
        is_empty = (len(patt_cmd) == 0)
        is_not_vec = (patt_cmd[-1].have_cmd(PattVecCmd.CMD_VECTOR) == False)
        if self.expanding_procs == False and (is_empty or is_not_vec):
            err_msg = "COMPILER ERROR: At least one Vector statement is required before Call statement!"
            raise Exception(err_msg)
        else:
            self.add_cmd_macro(PattVecCmd.CMD_CALL, self.curr_proc_name_call)

    def b_proc__pattern_statements__CALL_PROC_NAME(self, t):
        super().b_proc__pattern_statements__CALL_PROC_NAME(t)
        patt_cmd = self.proc2cmd[self.curr_proc_name]
        is_empty = (len(patt_cmd) == 0)
        is_not_vec = (patt_cmd[-1].have_cmd(PattVecCmd.CMD_VECTOR) == False)
        if self.expanding_procs == False and (is_empty or is_not_vec):
            err_msg = "COMPILER ERROR: At least one Vector statement is required before Call statement!"
            raise STILDumpCompilerException(-1, -1, err_msg)
        else:
            self.add_cmd_proc(PattVecCmd.CMD_CALL, self.curr_proc_name_call)

    def b_pattern__pattern_statements__CALL_MACRO_NAME(self, t):
        super().b_pattern__pattern_statements__CALL_MACRO_NAME(t)
        self.add_cmd_patt(PattVecCmd.CMD_MACRO, self.curr_macro_name_call)

    def b_macro__pattern_statements__CALL_MACRO_NAME(self, t):
        super().b_macro__pattern_statements__CALL_MACRO_NAME(t)
        self.add_cmd_macro(PattVecCmd.CMD_MACRO, self.curr_macro_name_call)

    def b_proc__pattern_statements__CALL_MACRO_NAME(self, t):
        super().b_proc__pattern_statements__CALL_MACRO_NAME(t)
        self.add_cmd_proc(PattVecCmd.CMD_MACRO, self.curr_macro_name_call)

    def b_pattern__pattern_statements__LOOP_COUNT(self, t, suppress_by_subclass = False):
        super().b_pattern__pattern_statements__LOOP_COUNT(t)
        
        if suppress_by_subclass == False:
            patt_cmd = self.patt2cmd[self.curr_pattern]
            if len(patt_cmd) == 0 or patt_cmd[-1].have_cmd(PattVecCmd.CMD_VECTOR) == False:
                err_msg = "COMPILER ERROR: At least one Vector statement is required before Loop block!"
                raise STILDumpCompilerException(-1, -1, err_msg)
            else:
                self.save_in_prev_cmd_patt(PattVecCmd.CMD_LOAD_LOOP_COUNTER, t.value)
                self.add_cmd_patt(PattVecCmd.CMD_START_LOOP)

    def b_macrodefs__pattern_statements__LOOP_COUNT(self, t, suppress_by_subclass = False):
        super().b_macrodefs__pattern_statements__LOOP_COUNT(t)

        if suppress_by_subclass == False:
            patt_cmd = self.macro2cmd[self.curr_macro_name]
            if len(patt_cmd) == 0 or patt_cmd[-1].have_cmd(PattVecCmd.CMD_VECTOR) == False:
                err_msg = "COMPILER ERROR: At least one Vector statement is required before Loop block!"
                raise STILDumpCompilerException(-1, -1, err_msg)
            else:
                self.save_in_prev_cmd_macro(PattVecCmd.CMD_LOAD_LOOP_COUNTER, t.value)
                self.add_cmd_macro(PattVecCmd.CMD_START_LOOP, t.value)

    def b_procedures__pattern_statements__LOOP_COUNT(self, t, suppress_by_subclass = False):
        super().b_procedures__pattern_statements__LOOP_COUNT(t)

        if suppress_by_subclass == False:
            patt_cmd = self.proc2cmd[self.curr_proc_name]
            if len(patt_cmd) == 0 or patt_cmd[-1].have_cmd(PattVecCmd.CMD_VECTOR) == False:
                err_msg = "COMPILER ERROR: At least one Vector statement is required before Loop block!"
                raise STILDumpCompilerException(-1, -1, err_msg)
            else:
                self.save_in_prev_cmd_proc(PattVecCmd.CMD_LOAD_LOOP_COUNTER, t.value)
                self.add_cmd_proc(PattVecCmd.CMD_START_LOOP, t.value)

    def b_pattern__pattern_statements__close_loop_block(self, t):
        super().b_pattern__pattern_statements__close_loop_block(t)
        self.save_in_prev_cmd_patt(PattVecCmd.CMD_STOP_LOOP)

    def b_macrodefs__pattern_statements__close_loop_block(self, t):
        super().b_macrodefs__pattern_statements__close_loop_block(t)
        self.save_in_prev_cmd_macro(PattVecCmd.CMD_STOP_LOOP)

    def b_procedures__pattern_statements__close_loop_block(self, t):
        super().b_procedures__pattern_statements__close_loop_block(t)
        self.save_in_prev_cmd_proc(PattVecCmd.CMD_STOP_LOOP)

    def b_pattern__pattern_statements__MATCHLOOP_COUNT(self, t, suppress_by_subclass = False):
        super().b_pattern__pattern_statements__MATCHLOOP_COUNT(t)

        if suppress_by_subclass == False:
            patt_cmd = self.patt2cmd[self.curr_pattern]
            if len(patt_cmd) == 0 or patt_cmd[-1].have_cmd(PattVecCmd.CMD_VECTOR) == False:
                err_msg = "COMPILER ERROR: At least one Vector statement is required before Loop block!"
                raise STILDumpCompilerException(-1, -1, err_msg)
            else:
                self.save_in_prev_cmd_patt(PattVecCmd.CMD_LOAD_MATCHLOOP_COUNTER, t.value)
                self.add_cmd_patt(PattVecCmd.CMD_START_MATCHLOOP)

    def b_macrodefs__pattern_statements__MATCHLOOP_COUNT(self, t, suppress_by_subclass = False):
        super().b_macrodefs__pattern_statements__MATCHLOOP_COUNT(t)
        
        if suppress_by_subclass == False:
            patt_cmd = self.macro2cmd[self.curr_macro_name]
            if len(patt_cmd) == 0 or patt_cmd[-1].have_cmd(PattVecCmd.CMD_VECTOR) == False:
                err_msg = "COMPILER ERROR: At least one Vector statement is required before Loop block!"
                raise STILDumpCompilerException(-1, -1, err_msg)
            else:
                self.save_in_prev_cmd_macro(PattVecCmd.CMD_LOAD_MATCHLOOP_COUNTER, t.value)
                self.add_cmd_macro(PattVecCmd.CMD_STRAT_MATCHLOOP, t.value)


    def b_procedures__pattern_statements__MATCHLOOP_COUNT(self, t, suppress_by_subclass = False):
        super().b_procedures__pattern_statements__MATCHLOOP_COUNT(t)

        if suppress_by_subclass == False:
            patt_cmd = self.proc2cmd[self.curr_proc_name]
            if len(patt_cmd) == 0 or patt_cmd[-1].have_cmd(PattVecCmd.CMD_VECTOR) == False:
                err_msg = "COMPILER ERROR: At least one Vector statement is required before Loop block!"
                raise STILDumpCompilerException(-1, -1, err_msg)
            else:
                self.save_in_prev_cmd_proc(PattVecCmd.CMD_LOAD_MATCHLOOP_COUNTER, t.value)
                self.add_cmd_proc(PattVecCmd.CMD_START_MATCHLOOP, t.value)

    def b_pattern__pattern_statements__MATCHLOOP_INF(self, t):
        super().b_pattern__pattern_statements__MATCHLOOP_INF(t)
        self.add_cmd_patt(PattVecCmd.CMD_START_MATCHLOOP, t.value)
        self.is_matchloop_inf = True

    def b_macrodefs__pattern_statements__MATCHLOOP_INF(self, t):
        super().b_macrodefs__pattern_statements__MATCHLOOP_INF(t)
        self.add_cmd_macro(PattVecCmd.CMD_START_MATCHLOOP, t.value)
        self.is_matchloop_inf = True

    def b_procedures__pattern_statements__MATCHLOOP_INF(self, t):
        super().b_procedures__pattern_statements__MATCHLOOP_INF(t)
        self.add_cmd_proc(PattVecCmd.CMD_START_MATCHLOOP, t.value)
        self.is_matchloop_inf = True

    def b_pattern__pattern_statements__close_matchloop_block(self, t):
        super().b_pattern__pattern_statements__close_matchloop_block(t)
        if self.is_matchloop_inf:
            self.save_in_prev_cmd_patt(PattVecCmd.CMD_STOP_INF_MATCHLOOP)
            self.is_matchloop_inf = False
        else:
            self.save_in_prev_cmd_patt(PattVecCmd.CMD_STOP_COUNT_MATCHLOOP)

    def b_macrodefs__pattern_statements__close_matchloop_block(self, t):
        super().b_macrodefs__pattern_statements__close_matchloop_block(t)
        if self.is_matchloop_inf:
            self.save_in_prev_cmd_macro(PattVecCmd.CMD_STOP_INF_MATCHLOOP)
            self.is_matchloop_inf = False
        else:
            self.save_in_prev_cmd_macro(PattVecCmd.CMD_STOP_COUNT_MATCHLOOP)

    def b_procedures__pattern_statements__close_matchloop_block(self, t):
        super().b_procedures__pattern_statements__close_matchloop_block(t)
        if self.is_matchloop_inf:
            self.save_in_prev_cmd_proc(PattVecCmd.CMD_STOP_INF_MATCHLOOP)
            self.is_matchloop_inf = False
        else:
            self.save_in_prev_cmd_proc(PattVecCmd.CMD_STOP_COUNT_MATCHLOOP)

    def b_pattern__pattern_statements__CLOSE_CALL_VECTOR_BLOCK(self, t):
        self.save_cmd_patt()

    def b_pattern__pattern_statements__PROC_CALL_END_STMT(self, t):
        self.save_cmd_patt()

    def b_pattern__pattern_statements__CLOSE_MACRO_VECTOR_BLOCK(self, t):
        self.save_cmd_patt()

    def b_pattern__pattern_statements__MACRO_CALL_END_STMT(self, t):
        self.save_cmd_patt()

    def b_pattern__pattern_statements__KEYWORD_IDDQ_TEST_POINT(self, t):
        self.add_cmd_patt(PattVecCmd.CMD_IDDQTESTPOINT, None)
        
    def b_macrodefs__pattern_statements__KEYWORD_IDDQ_TEST_POINT(self, t):
        self.add_cmd_macro(PattVecCmd.CMD_IDDQTESTPOINT, None)
        
    def b_procedures__pattern_statements__KEYWORD_IDDQ_TEST_POINT(self, t):
        self.add_cmd_proc(PattVecCmd.CMD_IDDQTESTPOINT, None)
        
    def b_pattern__pattern_statements__KEYWORD_STOP(self, t):
        self.save_in_prev_cmd_patt(PattVecCmd.CMD_STOP, None)

    def b_pattern__pattern_statements__KEYWORD_STOP_SC(self, t):
        self.save_in_prev_cmd_patt(PattVecCmd.CMD_STOP, None)

    def b_macrodefs__pattern_statements__KEYWORD_STOP(self, t):
        self.save_in_prev_cmd_macro(PattVecCmd.CMD_STOP, None)

    def b_macrodefs__pattern_statements__KEYWORD_STOP_SC(self, t):
        self.save_in_prev_cmd_macro(PattVecCmd.CMD_STOP, None)

    def b_procedures__pattern_statements__KEYWORD_STOP(self, t):
        self.save_in_prev_cmd_proc(PattVecCmd.CMD_STOP, None)
        
    def b_procedures__pattern_statements__KEYWORD_STOP_SC(self, t):
        self.save_in_prev_cmd_proc(PattVecCmd.CMD_STOP, None)

    def b_macrodefs__MACRO_NAME(self, t):

        super().b_macrodefs__MACRO_NAME(t)
        
        self.sgd2last_sig2wfc = {}
        self.sgd2before_subs_sig2wfc = {}
        
        self.macro2cmd[self.curr_macro_name] = []
        self.macro2signal_group_domains = {}

        for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
            patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
            # For every pattern burst block in the pattern exec block
            for patt_burst in patt_bursts:
    
                macro_domain = self.patt_burst2macro_domain[patt_burst]
                sig_group_domain = self.patt_burst2sig_groups_domain[patt_burst]
                
                if macro_domain in self.macro_domain2macro_names:
                    
                    macros = self.macro_domain2macro_names[macro_domain]
                
                    for macro in macros:
                        mdn = self.curr_macro_name
                        if mdn in self.macro2signal_group_domains:
                            signal_group_domains = self.macro2signal_group_domains[mdn]
                            if sig_group_domain not in signal_group_domains:
                                signal_group_domains.append(sig_group_domain)
                        else:
                            self.macro2signal_group_domains[mdn] = [sig_group_domain]
            
    def b_procedures__PROCEDURE_NAME(self, t):

        super().b_procedures__PROCEDURE_NAME(t)

        self.sgd2last_sig2wfc = {}
        self.sgd2before_subs_sig2wfc = {}
        
        self.proc2cmd[self.curr_proc_name] = []

        self.proc2signal_group_domains = {}

        for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
            patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
            # For every pattern burst block in the pattern exec block
            for patt_burst in patt_bursts:
    
                proc_domain = self.patt_burst2proc_domain[patt_burst]
                sig_group_domain = self.patt_burst2sig_groups_domain[patt_burst]
                
                if proc_domain in self.proc_domain2proc_names:
                    
                    procs = self.proc_domain2proc_names[proc_domain]
                
                    for proc in procs:
                        pdn = self.curr_proc_name
                        if pdn in self.proc2signal_group_domains:
                            signal_group_domains = self.proc2signal_group_domains[pdn]
                            if sig_group_domain not in signal_group_domains:
                                signal_group_domains.append(sig_group_domain)
                        else:
                            self.proc2signal_group_domains[pdn] = [sig_group_domain]

    def b_pattern__PATTERN_NAME(self, t):
        super().b_pattern__PATTERN_NAME(t)

        self.patt2cmd[self.curr_pattern] = []

    def dump_procedures(self):
        
        for pattern_exec_block in self.patt_exec_block2patt_burst.keys():

            time_domain = self.patt_exec_block2time_domain[pattern_exec_block]
            # time_domain = self.patt_exec_block2time_domain[pattern_exec_block]
            patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
            # For every pattern burst block in the pattern exec block
            for patt_burst in patt_bursts:

                proc_domain = self.patt_burst2proc_domain[patt_burst]
                macro_domain = self.patt_burst2macro_domain[patt_burst]
                sig_group_domain = self.patt_burst2sig_groups_domain[patt_burst]

                if proc_domain in self.proc_domain2proc_names:
                    procs = self.proc_domain2proc_names[proc_domain]
                    for proc in procs:
                        va = self.dump_procedure(
                            proc, sig_group_domain, proc_domain, macro_domain, time_domain
                        )
                        fpn = DomainUtils.get_full_name(proc_domain, proc)
                        self.proc2vas[fpn] = va

    def dump_procedure(self, proc, sgdomain, pdomain, mdomain, time_domain):

        #print("dump_procedure")
        
        fpn = DomainUtils.get_full_name(pdomain, proc)

        pd = DomainUtils.get_domain(pdomain, True)
        md = DomainUtils.get_domain(mdomain, True)
        sgd = DomainUtils.get_domain(sgdomain, True)

        fn = proc + "_" + pd + "_" + md + "_" + sgd + ".procedure_block"
        fn = fn.replace('"', "_")
         
        file = os.path.join(self.out_folder, fn)
        
        va = self.dump_patt_stmt_block(self.sgd_proc2sig2WFCs, self.proc2cmd , fpn, file, time_domain, sgdomain, pdomain, mdomain, False)
        return va 
    
    def dump_pattern_flow(self):
        
        if self.dump_data:
            # Processing pattern flow before the first pattern block
            file = os.path.join(self.out_folder, "pattern_blocks.flow")
    
            f = open(file, "w")
            f.write(
                "#FORMAT : pattern dump file name | pattern block name | td = Timing domain| sg = SignalsGroup domain | md = MacroDefs domain | pd = Procedures domain\n"
            )
    
            for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
    
                time_domain = self.patt_exec_block2time_domain[pattern_exec_block]
                patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
                # For every pattern burst block in the pattern exec block
                for patt_burst in patt_bursts:
    
                    pattern_list = self.patt_burst_block2pattern_blocks[patt_burst]
                    proc_domain = self.patt_burst2proc_domain[patt_burst]
                    macro_domain = self.patt_burst2macro_domain[patt_burst]
                    sig_group_domain = self.patt_burst2sig_groups_domain[patt_burst]
    
                    # For every pattern block in the pattern burst block
                    for patt in pattern_list:
                        td = DomainUtils.get_domain(time_domain, True)
                        sgd = DomainUtils.get_domain(sig_group_domain, True)
                        md = DomainUtils.get_domain(macro_domain, True)
                        pd = DomainUtils.get_domain(proc_domain, True)
                        ext = ".pattern_block"
    
                        fn = patt + "-" + td + "-" + sgd + "-" + md + "-" + pd + ext
    
                        rec = (
                            fn
                            + " | "
                            + patt
                            + " | td="
                            + td
                            + " | sgd="
                            + sgd
                            + " | md="
                            + md
                            + " | pd="
                            + pd
                            + "\n"
                        )
                        f.write(rec)
            f.close()

    def b_pattern__pattern_statements__CLOSE_VECTOR_BLOCK(self, t):
        
        #print("b_pattern__pattern_statements__CLOSE_VECTOR_BLOCK")

        # Current pattern block is not used in pattern burst
        if self.curr_pattern not in self.patt2sig_group_domain:
            super().b_pattern__pattern_statements__CLOSE_VECTOR_BLOCK(t)
            return

        #Ignore so far the Fixed and Condition statements
        if self.is_vector_stmt:
            self.save_cmd_patt()

            sig_groups = self.patt2sig_group_domain[self.curr_pattern]
            #print(f"sig_groups {sig_groups}")
            
            """
            Collecting WFC for all signals in the pattern's Vector statetment
            """
            for sgd in sig_groups:
                #print(f"sgd = {sgd} sig_groups = {sig_groups}")
                sgd_patt = DomainUtils.get_full_name(sgd, self.curr_pattern)
                if sgd_patt not in self.sgd_patt2sig2WFCs:
                    self.sgd_patt2sig2WFCs[sgd_patt] = {}
    
                si = 0
                total_wfc = 0
                for sigs in self.curr_sig_order:
                    sig2wfc = self.sgd_patt2sig2WFCs[sgd_patt]
                    if sigs in self.sig2type:
                        if sigs not in self.sgd_patt2sig2WFCs[sgd_patt]:
                            sig2wfc[sigs] = [self.curr_wfc_order[si]]
                        else:
                            sig2wfc[sigs].append(self.curr_wfc_order[si])
                        #print(f"signal {sigs} sig2wfc {sig2wfc}")
                        total_wfc += 1
                    else:
                        sig = DomainUtils.get_full_name(sgd, sigs)
                        if sig in self.signal_groups2signals:
                            wfci = 0
                            for s in self.signal_groups2signals[sig]:
                                # print(f"signal {s}")
                                if s not in self.sgd_patt2sig2WFCs[sgd_patt]:
                                    sig2wfc[s] = [self.curr_wfc_order[si][wfci]]
                                else:
                                    sig2wfc[s].append(self.curr_wfc_order[si][wfci])
                                #print(f"signal {s} sig2wfc {sig2wfc}")
                                wfci += 1
                            total_wfc += wfci
                        else:
                            sig = DomainUtils.get_name(sig)
                            err_msg = f"Can not find a signal/signal group '{sig}' in the pattern {self.curr_pattern}!"
                            raise Exception(err_msg)
                    si += 1
    
            if self.is_first_vector:
                self.patt2wfcs_number[self.curr_pattern] = total_wfc
            else:
                expected_wfc_number = self.patt2wfcs_number[self.curr_pattern]
                if total_wfc < expected_wfc_number:
                    #print(f"Delta vector expected_wfc_number {expected_wfc_number} != {total_wfc}")
                    """
                    Delta vector, not all WFC are defined
                    """
                    for sgd in sig_groups:
                        sgd_patt = DomainUtils.get_full_name(sgd, self.curr_pattern)
                        sigs2wfc = self.sgd_patt2sig2WFCs[sgd_patt]
                        max_wfc = 0
                        for sig in sigs2wfc.keys(): 
                            ml = len(sigs2wfc[sig])
                            if ml > max_wfc:
                                max_wfc = ml
                        for sig in sigs2wfc.keys():
                            if len(sigs2wfc[sig]) != max_wfc:
                                last_wfc = sigs2wfc[sig][-1]
                                if last_wfc == "#" or last_wfc == "%":
                                    before_subs_sig2wfc = self.sgd2before_subs_sig2wfc[sgd]
                                    if sig in before_subs_sig2wfc:
                                        wfc = before_subs_sig2wfc[sig]
                                        sigs2wfc[sig].append(wfc)
                                    else:
                                        err_msg = f"Missing WFC for a signal {sig} after a vector with '#' or '%'!"
                                        raise Exception(err_msg)
                                        
                                else:
                                    sigs2wfc[sig].append(last_wfc)
        
        if self.is_first_vector and (self.is_condition_stmt or self.is_vector_stmt):
            sig_groups = self.patt2sig_group_domain[self.curr_pattern]

            """
            Collecting all WFC per signals in the first Condition or Vector statetment
            """
    
            for sgd in sig_groups:
                sgd_patt = DomainUtils.get_full_name(sgd, self.curr_pattern)
                if sgd_patt not in self.sgd_patt2signals:
                    self.sgd_patt2signals[sgd_patt] = []
    
                sig_idx = 0
                for sigs in self.curr_sig_order:
                    all_found_sigs = self.sgd_patt2signals[sgd_patt]
                    if sigs in self.sig2type:
                        if sigs not in all_found_sigs:
                            all_found_sigs.append(sigs)
                        sgd_p_s = sgd_patt + "::" + sigs
                        self.sgd_patt_sig2defWFC[sgd_p_s] = self.curr_wfc_order[sig_idx]
                        
                    else:
                        sig = DomainUtils.get_full_name(sgd, sigs)
                        if sig in self.signal_groups2signals:
                            wfc_idx = 0
                            for s in self.signal_groups2signals[sig]:
                                if s not in all_found_sigs:
                                    all_found_sigs.append(s)
                                sgd_p_s = sgd_patt + "::" + s
                                self.sgd_patt_sig2defWFC[sgd_p_s] = self.curr_wfc_order[sig_idx][wfc_idx]
                            wfc_idx += 1
                        else:
                            sig = DomainUtils.get_name(sig)
                            err_msg = f"Can not find a signal/signal group '{sig}' in the pattern {self.curr_pattern}!"
                            raise Exception(err_msg)
                    sig_idx += 1
            for sgd_patt in self.sgd_patt2signals.keys():
                sigs = self.sgd_patt2signals[sgd_patt]
                for sig in sigs:
                    sgd_p_s = sgd_patt + "::" + sig
                    wfc = self.sgd_patt_sig2defWFC[sgd_p_s]
                    if sig not in self.signals_order:
                        self.signals_order.append(sig)
        
        super().b_pattern__pattern_statements__CLOSE_VECTOR_BLOCK(t)
        self.is_fixed_stmt = False
        self.is_condition_stmt = False
        self.is_vector_stmt = False

    def dump_pattern_blocks(self):
        
        pattern_blocks_va = 0
        proc_blocks_va = 0

        for pattern_exec_block in self.patt_exec_block2patt_burst.keys():
            #print(f"pattern_exec_block {pattern_exec_block}")
            time_domain = self.patt_exec_block2time_domain[pattern_exec_block]
            patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
            # For every pattern burst block in the pattern exec block
            for patt_burst in patt_bursts:
                #print(f"patt_burst {patt_burst}")

                pattern_list = self.patt_burst_block2pattern_blocks[patt_burst]
                proc_domain = self.patt_burst2proc_domain[patt_burst]
                macro_domain = self.patt_burst2macro_domain[patt_burst]
                sig_group_domain = self.patt_burst2sig_groups_domain[patt_burst]

                #print(f"pattern_list {pattern_list}")

                # For every pattern block in the pattern burst block
                for pattern in pattern_list:
                    td = DomainUtils.get_domain(time_domain, True)
                    sgd = DomainUtils.get_domain(sig_group_domain, True)
                    md = DomainUtils.get_domain(macro_domain, True)
                    pd = DomainUtils.get_domain(proc_domain, True)
                    ext = ".pattern_block"
                    fn = pattern + "-" + td + "-" + sgd + "-" + md + "-" + pd + ext
                    fn = fn.replace('"', "_")
                    patt_va, proc_va = self.dump_patt_stmt_block(self.sgd_patt2sig2WFCs, self.patt2cmd,  pattern, fn, time_domain, sig_group_domain, macro_domain, proc_domain, True)
                    pattern_blocks_va += patt_va
                    proc_blocks_va += proc_va
                    
        return pattern_blocks_va, proc_blocks_va
    
    def initial_fill_data(self, data, va_length):
        sep= ['|']*(va_length)
        empty_va = ['-']*(va_length)
        # [0]     VA - vector addresses
        data.append(range(va_length))
        # [1]     | separator 
        data.append(sep)
        # [2]     VAP - pattern only vector addresses
        data.append(empty_va)
        # [3]     | separator 
        data.append(sep)
        # [4]     VAMP - macro/proc only vector addresses
        data.append(empty_va)
        # [5]     | separator 
        data.append(sep)
        return data

    def wfc_substitution(self, hash_info, arg_sig2wfc, sig2wfcs, signals, sig_group_domain, cond_sig2wfc):
        """
        Replacing # and % with WFC from Macro/Procedure invocation according 
        the STIL standard.
        """
        # Key   - the signal name
        # value - wfc length of signals which are called in the Macro/Proc
        arg_wfcs_length = {}
        # Key - the signal name, value - number of pre-shift # per this signal
        pre_shift = {}
        # Key - the signal name, value - number of post-shift # per this signal
        post_shift = {}
        # List with signals which have # in any vector address
        signals_with_hash = []
        # List with signals which have # in a shift block
        signals_with_shift_hash = []
        # Key   - the signal name
        # value - shift cycles = arg_wfcs_length - (pre_shift + post_shift)
        shift_cycles = {}
        # Maximum value in shift_cycles
        max_scan_shift = 0        
        # Key   - the signal name
        # value - padding WFC count = max_scan_shift - shift_cycles
        padding_len = {}
        # List with input type signals which have to be pre padded
        prepad_sig = []
        # List with output type signals which have to be post padded
        postpad_sig = []
           
        for sig in signals:
            if sig in arg_sig2wfc:
                wfcs = arg_sig2wfc[sig]
                arg_wfcs_length[sig] = len(wfcs)
                #print(f"sig {sig} len {len(wfcs)}")
                
        # in the case of signals group part of argument WFC
        new_sig2wfc = {}
        for sg in arg_sig2wfc:
            fsg = DomainUtils.get_full_name(sig_group_domain, sg)
            if fsg in self.signal_groups2signals:
                signals_group = self.signal_groups2signals[fsg]
                si = 0
                for signal in signals_group:
                    wfcs = arg_sig2wfc[sg][si]
                    arg_wfcs_length[signal] = len(wfcs)
                    new_sig2wfc[signal] = [wfcs]
                    si += 1
        for sig in new_sig2wfc:
            arg_sig2wfc[sig] = new_sig2wfc[sig]
        
        for hi in hash_info:

            if hi.pos_rel_shift == HashInfo.POS_BEFORE_SHIFT:
                
                if self.expanding_procs == False:
                    err_msg = "When option expanding_procs is False, '#' before Shift block is not allowed!"
                    raise STILDumpCompilerException(-1, -1, err_msg)
                
                if hi.sig_ref in self.sig2type:
                    if hi.sig_ref in pre_shift:
                        pre_shift[hi.sig_ref] += 1
                    else:
                        pre_shift[hi.sig_ref] = 1
                    if hi.sig_ref not in signals_with_hash:
                        signals_with_hash.append(hi.sig_ref)
                else:
                    sg = DomainUtils.get_full_name(sig_group_domain, hi.sig_ref)
                    sig_group = self.signal_groups2signals[sg]
                    for pos in hi.pos:
                        sig = sig_group[pos]
                        if sig in pre_shift:
                            pre_shift[sig] += 1
                        else:
                            pre_shift[sig] = 1
                        if sig not in signals_with_hash:
                            signals_with_hash.append(sig)
                            
            elif hi.pos_rel_shift == HashInfo.POS_AFTER_SHIFT:
                
                if self.expanding_procs == False:
                    err_msg = "When option expanding_procs is False, '#' after Shift block is not allowed!"
                    raise STILDumpCompilerException(-1, -1, err_msg)

                if hi.sig_ref in self.sig2type:
                    if hi.sig_ref in post_shift:
                        post_shift[hi.sig_ref] += 1
                    else:
                        post_shift[hi.sig_ref] = 1
                    if hi.sig_ref not in signals_with_hash:
                        signals_with_hash.append(hi.sig_ref)
                else:
                    sg = DomainUtils.get_full_name(sig_group_domain, hi.sig_ref)
                    sig_group = self.signal_groups2signals[sg]
                    for pos in hi.pos:
                        sig = sig_group[pos]
                        if sig in post_shift:
                            post_shift[sig] += 1
                        else:
                            post_shift[sig] = 1
                        if sig not in signals_with_hash:
                            signals_with_hash.append(sig)
                            
            elif hi.pos_rel_shift == HashInfo.POS_SHIFT:
                
                if hi.sig_ref in self.sig2type:
                    if hi.sig_ref not in signals_with_hash:
                        signals_with_hash.append(hi.sig_ref)
                    if hi.sig_ref not in signals_with_shift_hash:
                        signals_with_shift_hash.append(hi.sig_ref)
                        
                else:
                    sg = DomainUtils.get_full_name(sig_group_domain, hi.sig_ref)
                    sig_group = self.signal_groups2signals[sg]
                    for pos in hi.pos:
                        sig = sig_group[pos]
                        if sig not in signals_with_hash:
                            signals_with_hash.append(sig)
                        if sig not in signals_with_shift_hash:
                            signals_with_shift_hash.append(sig)
        
        for sig in signals_with_hash:
            if sig not in arg_sig2wfc:
                arg_sig2wfc[sig] = ""
        
        for sig in signals:
            
            l = 0
            if sig in arg_wfcs_length:
                l = arg_wfcs_length[sig]
                
            pre_shift_count = 0
            if sig in pre_shift:
                pre_shift_count = pre_shift[sig]
                
            post_shift_count = 0
            if sig in post_shift:
                post_shift_count = post_shift[sig]

            if sig in signals_with_shift_hash:
               
                shift_cycles[sig] = l - (pre_shift_count + post_shift_count)
                if shift_cycles[sig] > max_scan_shift:
                    max_scan_shift = shift_cycles[sig]
            
        # Padding WFC count = max_scan_shift - shift_cycles
        for sig in signals:
            if sig in shift_cycles:
                pad = max_scan_shift - shift_cycles[sig]
                if pad > 0:
                    padding_len[sig] = pad
                    if self.sig2type[sig] == "In":
                        prepad_sig.append(sig)
                        #print(f"pre padding for signal {sig} : {pad}")
                    elif self.sig2type[sig] == "Out":
                        postpad_sig.append(sig)
                        #print(f"post padding for signal {sig} : {pad}")
                    elif sig in self.scanin_sigs:
                        postpad_sig.append(sig)
                    elif sig in self.scanout_sigs:
                        postpad_sig.append(sig)
                    else:
                        err_msg = "COMPILER ERROR: padding is unknown for none unidirectional signal"
                        raise STILDumpCompilerException(-1, -1, err_msg)
                        
        # Pre-padding
        for sig in prepad_sig:
            
            if sig not in sig2wfcs:
                continue
            
            wfcs = sig2wfcs[sig]
            # Get the last defined WFC before the #
            indx = -1;
            prepad_value = None
            for wfc in wfcs:
                if wfc == "#":
                    if indx > -1:
                        prepad_value = wfcs[indx]
                    break
                else:
                    indx += 1
                    
            if prepad_value == None:
                if cond_sig2wfc != None:
                    if sig in cond_sig2wfc:
                        prepad_value = cond_sig2wfc[sig][0]
                    
            if prepad_value == None:
                err_msg = "COMPILER ERROR: Can not find pre-padding value in the vector before the first #"          
                raise STILDumpCompilerException(-1, -1, err_msg)
            else:
                pad_count = padding_len[sig]
                padding_data = pad_count*prepad_value
                arg_sig2wfc[sig] = padding_data + arg_sig2wfc[sig]

        # Post-padding
        for sig in postpad_sig:
            
            if sig not in sig2wfcs:
                continue
            
            wfcs = sig2wfcs[sig]
            # Get the last defined WFC before the #
            indx = -1;
            postpad_value = None
            for wfc in wfcs:
                if wfc == "#":
                    if indx > -1:
                        postpad_value = wfcs[indx]
                    break
                else:
                    indx += 1

            if postpad_value == None:
                if sig in cond_sig2wfc:
                    postpad_value = cond_sig2wfc[sig][0]

            if postpad_value == None:
                err_msg = "COMPILER ERROR: Can not find post-padding value in the vector before the first #"        
                raise STILDumpCompilerException(-1, -1, err_msg)
            else:
                pad_count = padding_len[sig]
                padding_data = pad_count*postpad_value
                arg_sig2wfc[sig] = arg_sig2wfc[sig] + padding_data
                #print(f"post padding arg_sig2wfc[{sig}] {arg_sig2wfc[sig]}")

    
        # Key is the signal name
        # Value is a list with the WFC of this signal per each VA
        new_sig2wfcs = {}
        # Key is the signal name
        # Value is a list with scan chain memory filled with WFC
        scan_mem = {}
        
        shift_pos = -1

        # Processing signals which have # in shift block        
        for sig in signals:
            
            if sig not in sig2wfcs:
                continue
            
            wfcs = sig2wfcs[sig]
            
            new_vectors = 0
            
            if sig in arg_sig2wfc and sig in signals_with_shift_hash:
                #print(f" signals_with_shift_hash {sig}")
                new_sig2wfcs[sig] = []
                pre = 0
                if sig in pre_shift:
                    pre = pre_shift[sig]
                post = 0
                if sig in post_shift:
                    post = post_shift[sig]
                is_shift_done = False
                arg_indx = 0
                for i in range(len(wfcs)):
                    if wfcs[i] == '#':
                        if pre > 0:
                            new_sig2wfcs[sig].append(arg_sig2wfc[sig][arg_indx])
                            arg_indx += 1
                            pre -= 1
                        elif is_shift_done == False:
                            shift_pos = i
                            is_hash_added = False
                            for ii in range(max_scan_shift):
                                wfc = arg_sig2wfc[sig][arg_indx]
                                if self.is_scan_mem_available:
                                    if is_hash_added == False:
                                        is_hash_added = True
                                        new_sig2wfcs[sig].append("#")
                                    if sig not in scan_mem:
                                        scan_mem[sig] = []
                                    scan_mem[sig].append(wfc)
                                else:
                                    new_sig2wfcs[sig].append(wfc)
                                    new_vectors += 1
                                arg_indx += 1
                            is_shift_done = True
                        elif post > 0:
                            new_sig2wfcs[sig].append(arg_sig2wfc[sig][arg_indx])
                            arg_indx += 1
                            post -= 1
                    else:
                        new_sig2wfcs[sig].append(wfcs[i])

        # Processing signals which have #, but not in the shift block        
        for sig in signals:
            
            if sig not in sig2wfcs:
                continue
            
            wfcs = sig2wfcs[sig]
            
            if sig in arg_sig2wfc and \
                sig not in signals_with_shift_hash and \
                sig in signals_with_hash:
                    
                new_sig2wfcs[sig] = []
                pre = 0
                if sig in pre_shift:
                    pre = pre_shift[sig]
                post = 0
                if sig in post_shift:
                    post = post_shift[sig]
                arg_indx = 0
                
                for i in range(len(wfcs)):
                    if wfcs[i] == '#':
                        if pre > 0:
                            new_sig2wfcs[sig].append(arg_sig2wfc[sig][arg_indx])
                            arg_indx += 1
                            pre -= 1
                        elif post > 0:
                            new_sig2wfcs[sig].append(arg_sig2wfc[sig][arg_indx])
                            arg_indx += 1
                            post -= 1
                    elif i == shift_pos:
                        for ii in range(max_scan_shift):
                            new_sig2wfcs[sig].append(wfcs[i])
                    else:
                        if wfcs[i] == '?':
                            if self.expanding_procs == False:
                                # Issue #13
                                err_msg = "COMPILER ERROR: Delta vector after # is not supported in none expanding procedure mode"
                                raise STILDumpCompilerException(-1, -1, err_msg)
                            else:
                                w = new_sig2wfcs[sig]
                                new_sig2wfcs[sig].append(w[-1])
                        else:
                            new_sig2wfcs[sig].append(wfcs[i])
                        
        max_wfc_len = 0
                        
        # Processing signal without #
        for sig in signals:
            
            if sig not in sig2wfcs:
                continue
            
            wfcs = sig2wfcs[sig]
            
            if sig not in signals_with_hash:

                if len(wfcs) > max_wfc_len:
                    max_wfc_len = len(wfcs)

                new_sig2wfcs[sig] = []
                for i in range(len(wfcs)):
                    if i == shift_pos:
                        r = wfcs[i]
                        if self.is_scan_mem_available:
                            new_sig2wfcs[sig].append(r)
                        else:
                            for ii in range(max_scan_shift):
                                new_sig2wfcs[sig].append(r)
                    else:
                        new_sig2wfcs[sig].append(wfcs[i])

        # Processing signals which have % 
        for sig in signals:
            
            if sig not in sig2wfcs:
                continue
            
            wfcs = new_sig2wfcs[sig]
            if "%" in wfcs:
                subs_wfc = arg_sig2wfc[sig]
                if len(subs_wfc) == 1:
                    wfcs = [wfc.replace('%', subs_wfc) for wfc in wfcs]
                    new_sig2wfcs[sig] = wfcs
                else:
                    err_msg = "COMPILER ERROR: Only 1 WFC expected for % substitution of the signal {sig} "
                    raise STILDumpCompilerException(-1, -1, err_msg)
        
        return new_sig2wfcs, scan_mem, new_vectors

    def dump_patt_stmt_block(
        self, sgd_patt2sig2WFCs, block2cmd, block_name, file_name, timing, sig_group_domain, macro_domain, proc_domain, is_pattern_block
    ):
        
        '''
        This function will "compile" the input block2cmd (which can be Pattern or Procedure block)
        into text file
        '''
        
        self.data = []

        self.scan_data = []
        
        # Collect WFC data from sgd_patt2sig2WFCs
        # sgd_patt2sig2WFCs can be either pattern or procedure block
        signals = []
        sep = []
        sgd_patt = DomainUtils.get_full_name(sig_group_domain, block_name)
        
        sig2patt_wfcs = {}
        
        if len(self.signals_order) > 0:
            signals = self.signals_order
            

        # It is posible to have pattern block without any vector,
        # for example : only calls to macros and procedures
        if len(sgd_patt2sig2WFCs) > 0:
            sig2patt_wfcs = sgd_patt2sig2WFCs[sgd_patt]
            
            # ToDo: add signals which are in the Condition stmt, but not in the
            # vectors
            # from sgd_patt2signals and sgd_patt_sig2defWFC
                    
            va_count = 0
            for sig in sig2patt_wfcs:
                if sig not in signals:
                    signals.append(sig)
                wfcs = sig2patt_wfcs[sig]
                va_count = len(wfcs)
                #print(f"sig {sig} {wfcs} va_count {va_count}\n")
                if len(self.data) == 0:
                    self.data = self.initial_fill_data(self.data, va_count)
                # Adding WFCs data for the current signal
                self.data.append(wfcs)
            self.data.append(sep)
        
        # Vector address in the case of expanding Macro or Procedure blocks
        exp_va = 0
        vamp = []
        vap = []
        vap_c = 0
        cmds = block2cmd[block_name]
        # Key is the signal name
        # Value is a list with scan chain memory filled with WFC
        block_scan_mem = {}
        
        proc_va = 0
        
        va_cmd = []
        
        last_start_loop_va = -1
        
        for vec_cmds in cmds:
            #print("\tNEW COMMAND\n")
            
            va_cmd.append(vec_cmds)
                                    
            for cmd in vec_cmds.get_cmd_ids():
                cmd_name = PattVecCmd.get_cmd_name(cmd)
                #print(f"\t\tCMD {cmd_name} | VALUE {vec_cmds.get_value(cmd)}\n")
                
                if cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_VECTOR]:
                    #print("vector cmd")
                    vamp.append("-")
                    vap.append(vap_c)
                    vap_c += 1
                    exp_va += 1 
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_WFT]:
                    pass
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_LABEL]:
                    va_cmd[-1].set_value(PattVecCmd.CMD_LABEL, vec_cmds.get_value(cmd))
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_LOAD_LOOP_COUNTER]:
                    pass
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_LOAD_MATCHLOOP_COUNTER]:
                    pass
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_START_LOOP]:
                    if last_start_loop_va != -1:
                        int_err_msg = "COMPILER ERROR : nested loop block detected"
                        raise STILDumpCompilerException(-1, -1, int_err_msg)
                    else:
                        last_start_loop_va = exp_va
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP_LOOP]:
                    jmp = exp_va - last_start_loop_va - 1
                    va_cmd[-1].set_value(PattVecCmd.CMD_STOP_LOOP, jmp)
                    last_start_loop_va = -1
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_START_MATCHLOOP]:
                    value = vec_cmds.get_value(cmd)

                    if last_start_loop_va != -1:
                        int_err_msg = "COMPILER ERROR : nested loop block detected"
                        raise STILDumpCompilerException(-1, -1, int_err_msg)
                    else:
                        last_start_loop_va = exp_va

                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP_COUNT_MATCHLOOP]:
                    jmp = exp_va - last_start_loop_va - 1
                    va_cmd[-1].set_value(PattVecCmd.CMD_STOP_COUNT_MATCHLOOP, jmp)
                    last_start_loop_va = -1
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP_INF_MATCHLOOP]:
                    jmp = exp_va - last_start_loop_va - 1
                    va_cmd[-1].set_value(PattVecCmd.CMD_STOP_INF_MATCHLOOP, jmp)
                    last_start_loop_va = -1
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_IDDQTESTPOINT]:
                    pass
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_BREAKPOINT]:
                    pass
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_GOTO]:
                    pass
                    va_cmd[-1].set_value(PattVecCmd.CMD_GOTO, vec_cmds.get_value(cmd))
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP]:
                    pass
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_START_SHIFT]:
                    pass
                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP_SHIFT]:
                    pass
                elif cmd_name ==  PattVecCmd.cmds[PattVecCmd.CMD_MACRO]:
                    # Macro blocks are always expanding into the pattern block

                    macro_va_length = 0

                    fmn = DomainUtils.get_full_name(macro_domain, vec_cmds.get_value(cmd))

                    sgd_macro = sig_group_domain + "::" + fmn
                    sig2macro_wfcs = self.sgd_macro2sig2WFCs[sgd_macro]

                    if len(signals) == 0:
                        for sig in sig2macro_wfcs:
                            if sig in self.sig2type:
                                signals.append(sig)

                    is_new_data_created = False
                    if len(sig2patt_wfcs) == 0:
                        for sig in signals:
                            if sig in sig2macro_wfcs:
                                wfcs = sig2macro_wfcs[sig]
                                va_count = len(wfcs)
                            else:
                                wfcs = list(sig2macro_wfcs.keys())[0]
                                va_count = len(wfcs)
                            if len(self.data) == 0:
                                self.data = self.initial_fill_data(self.data, va_count)
                                is_new_data_created = True
                            if is_new_data_created:
                                wfc = [] 
                                self.data.append(wfc)
                        if is_new_data_created:
                            self.data.append(sep)
                    
                    # WFC data substitution
                    if fmn in self.macro_for_subs:
                        arg_sig2wfc = vec_cmds.get_props(PattVecCmd.CMD_MACRO)
                        # TODO : to check STIL standard what should be done
                        if arg_sig2wfc != None:
                            hash_info = self.hashinfo[fmn]
                            cond_sig2wfc = None
                            if fmn in self.macro2first_cond:
                                cond_sig2wfc = self.macro2first_cond[fmn]
                            exp_sig2wfc, scan_mem, new_vectors = self.wfc_substitution(
                                                    hash_info, 
                                                    arg_sig2wfc, 
                                                    sig2macro_wfcs, 
                                                    signals, 
                                                    sig_group_domain,
                                                    cond_sig2wfc)

                            if self.is_scan_mem_available:
                                sd = {}
                                for sig in scan_mem:
                                    if sig not in block_scan_mem:
                                        block_scan_mem[sig] = []
                                    block_scan_mem[sig].extend(scan_mem[sig])
                                    scan_chain = ''
                                    for wfc in scan_mem[sig]:
                                        scan_chain += wfc
                                    if len(scan_chain) > 0:
                                        sd[sig] = scan_chain
                                if len(sd) > 0:
                                    self.scan_data.append(sd)
                            else:
                                pass
                            
                            is_first = True
                            sig_index = 6
                            exp = 0
                            for sig in signals:
                                if sig in sig2macro_wfcs:
                                    wfcs = sig2macro_wfcs[sig]
                                if sig in exp_sig2wfc:
                                    wfcs = exp_sig2wfc[sig]
                                    macro_va_length = len(wfcs)
                                    # adding WFC for the next signal
                                    self.data[sig_index][exp_va:exp_va] = wfcs
        
                                    if is_first:
                                        exp += macro_va_length
                                        is_first = False
                                else:
                                    # For signal not in the expansion, takes the latest WFC (issue #13)
                                    last_wfc = self.data[sig_index][-1]
                                    self.data[sig_index][exp_va:exp_va] = last_wfc
                                sig_index += 1
                                
                            shift_sig_wo_data = []
                            if len(scan_mem) == 0:
                                for sig in exp_sig2wfc:
                                    wfcs = exp_sig2wfc[sig]
                                    if len(wfcs) == 0:
                                        shift_sig_wo_data.append(sig)

                            if len(shift_sig_wo_data) > 0:
                                print("WARNING: The following signals in Shift block do not have data in the call of the macro:")
                                for s in shift_sig_wo_data:
                                    print(f"         {s}")
                                macro_va_length = 1

                                sig_index = 6
                                for sig in signals:
                                    if sig in shift_sig_wo_data:
                                        wfcs = self.data[sig_index]
                                        last_wfc = self.data[sig_index][-1]
                                        self.data[sig_index][exp_va:exp_va] = last_wfc
                                    sig_index += 1

                            exp_va += exp
                        else:
                            # Macro call without data, but with shift inside the macro
                            exp_va += 1
                            macro_va_length = 1

                            sig_index = 6
                            for sig in signals:
                                last_wfc = self.data[sig_index][-1]
                                self.data[sig_index][exp_va:exp_va] = last_wfc
                                sig_index += 1

                    else:
                        is_first = True
                        sig_index = 6
                        exp = 0
                        for sig in signals:
                            if sig in sig2macro_wfcs:
                                wfcs = sig2macro_wfcs[sig]
                                macro_va_length = len(wfcs)
                                # adding WFC for the next signal
                                self.data[sig_index][exp_va:exp_va] = wfcs
    
                                if is_first:
                                    exp += macro_va_length
                                    is_first = False
                            sig_index += 1
                        exp_va += exp
                            
                    # In the case when the last command does not have Vector statement
                    # like macro or procedure invocation we have to "transfer" the new command to the old command
                    add_to_last_cmd = False
                    last_cmd = va_cmd[-1]
                    if last_cmd.have_cmd(PattVecCmd.CMD_VECTOR) == False:
                        add_to_last_cmd = True
                        
                    macro_commands = self.macro2cmd[fmn]
                    last_start_loop_va = -1
                    m_va = 0
                    for macro_command in macro_commands:
                        if add_to_last_cmd:
                            cids = macro_command.get_cmd_ids()
                            for cid in cids: 
                                value = macro_command.get_value(cid)
                                last_cmd.add_cmd(cid, value)
                            add_to_last_cmd = False
                        else:
                            va_cmd.append(macro_command)
                            for cmd in macro_command.get_cmd_ids():
                                cmd_name = PattVecCmd.get_cmd_name(cmd)
                                if cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_WFT]:
                                    pass
                                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_START_LOOP]:
                                    last_start_loop_va = m_va
                                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP_LOOP]:
                                    jmp = m_va - last_start_loop_va - 1
                                    va_cmd[-1].set_value(PattVecCmd.CMD_STOP_LOOP, jmp)
                                    last_start_loop_va = -1
                                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_START_MATCHLOOP]:
                                    last_start_loop_va = m_va
                                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP_COUNT_MATCHLOOP]:
                                    jmp = m_va - last_start_loop_va - 1
                                    va_cmd[-1].set_value(PattVecCmd.CMD_STOP_COUNT_MATCHLOOP, jmp)
                                    last_start_loop_va = -1
                                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP_INF_MATCHLOOP]:
                                    jmp = m_va - last_start_loop_va - 1
                                    va_cmd[-1].set_value(PattVecCmd.CMD_STOP_INF_MATCHLOOP, jmp)
                                    last_start_loop_va = -1
                                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_GOTO]:
                                    va_cmd[-1].set_value(PattVecCmd.CMD_GOTO, macro_command.get_value(cmd))
                                elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_LABEL]:
                                    va_cmd[-1].set_value(PattVecCmd.CMD_LABEL, macro_command.get_value(cmd))
                            m_va += 1
                            
                    for i in range(macro_va_length):
                        vap.append("-")
                        vamp.append(i)

                elif cmd_name ==  PattVecCmd.cmds[PattVecCmd.CMD_CALL]:

                    proc_va_length = 0

                    fpn = DomainUtils.get_full_name(proc_domain, vec_cmds.get_value(cmd))

                    if self.expanding_procs:
                        
                        sgd_proc = sig_group_domain + "::" + fpn
                        sig2proc_wfcs = self.sgd_proc2sig2WFCs[sgd_proc]

                        if len(signals) == 0:
                            for sig in sig2proc_wfcs:
                                if sig in self.sig2type:
                                    signals.append(sig)

                        is_new_data_created = False
                        if len(sig2patt_wfcs) == 0:
                            for sig in signals:
                                if sig in sig2proc_wfcs:
                                    wfcs = sig2proc_wfcs[sig]
                                    va_count = len(wfcs)
                                else:
                                    wfcs = list(sig2proc_wfcs.keys())[0]
                                    va_count = len(wfcs)
                                if len(self.data) == 0:
                                    self.data = self.initial_fill_data(self.data, va_count)
                                    is_new_data_created = True
                                if is_new_data_created:
                                    wfc = [] 
                                    self.data.append(wfc)
                            if is_new_data_created:
                                self.data.append(sep)
                        
                        # WFC data substitution
                        is_wfc_subs = False
                        scan_len = 0
                        if fpn in self.proc_for_subs:
                            arg_sig2wfc = vec_cmds.get_props(PattVecCmd.CMD_CALL)
                            # TODO : to check STIL standard what should be done
                            if arg_sig2wfc != None:
                                hash_info = self.hashinfo[fpn]
                                cond_sig2wfc = None
                                if fpn in self.proc2first_cond:
                                    cond_sig2wfc = self.proc2first_cond[fpn]
                                exp_sig2wfc, scan_mem, new_vectors = self.wfc_substitution(
                                                        hash_info, 
                                                        arg_sig2wfc, 
                                                        sig2proc_wfcs, 
                                                        signals, 
                                                        sig_group_domain,
                                                        cond_sig2wfc)

                                if self.is_scan_mem_available:
                                    sd = {}
                                    for sig in scan_mem:
                                        if sig not in block_scan_mem:
                                            block_scan_mem[sig] = []
                                        block_scan_mem[sig].extend(scan_mem[sig])
                                        scan_chain = ''
                                        for wfc in scan_mem[sig]:
                                            scan_chain += wfc
                                        if len(scan_chain) > 0:
                                            sd[sig] = scan_chain
                                    if len(sd) > 0:
                                        self.scan_data.append(sd)
                                else:
                                    is_wfc_subs = True
                                
                                is_first = True
                                sig_index = 6
                                exp = 0
                                for sig in signals:
                                    if sig in sig2proc_wfcs:
                                        wfcs = sig2proc_wfcs[sig]
                                    if sig in exp_sig2wfc:
                                        wfcs = exp_sig2wfc[sig]
                                        proc_va_length = len(wfcs)
                                        # adding WFC for the next signal
                                        self.data[sig_index][exp_va:exp_va] = wfcs
            
                                        if is_first:
                                            exp += proc_va_length
                                            scan_len = len(wfcs)
                                            is_first = False
                                    else:
                                        # For signal not in the expansion, takes the latest WFC (issue #13)
                                        last_wfc = self.data[sig_index][-1]
                                        self.data[sig_index][exp_va:exp_va] = last_wfc
                                    sig_index += 1

                                shift_sig_wo_data = []
                                if len(scan_mem) == 0:
                                    for sig in exp_sig2wfc:
                                        wfcs = exp_sig2wfc[sig]
                                        if len(wfcs) == 0:
                                            shift_sig_wo_data.append(sig)
    
                                if len(shift_sig_wo_data) > 0:
                                    print("WARNING: The following signals in Shift block do not have data in the call of the procedure:")
                                    for s in shift_sig_wo_data:
                                        print(f"         {s}")
                                    proc_va_length = 1

                                    sig_index = 6
                                    for sig in signals:
                                        if sig in shift_sig_wo_data:
                                            wfcs = self.data[sig_index]
                                            last_wfc = self.data[sig_index][-1]
                                            self.data[sig_index][exp_va:exp_va] = last_wfc
                                        sig_index += 1

                                exp_va += exp
                            else:
                                # Procedure call without data, but with shift inside the procedure
                                exp_va += 1
                                proc_va_length = 1

                                sig_index = 6
                                for sig in signals:
                                    last_wfc = self.data[sig_index][-1]
                                    self.data[sig_index][exp_va:exp_va] = last_wfc
                                    sig_index += 1
                                                                
                        else:
                            is_first = True
                            sig_index = 6
                            exp = 0
                            for sig in signals:
                                if sig in sig2proc_wfcs:
                                    wfcs = sig2proc_wfcs[sig]
                                    proc_va_length = len(wfcs)
                                    # adding WFC for the next signal
                                    self.data[sig_index][exp_va:exp_va] = wfcs
        
                                    if is_first:
                                        exp += proc_va_length
                                        is_first = False
                                sig_index += 1
                            exp_va += exp


                        # In the case when the last command does not have Vector statement
                        # like macro or procedure invocation we have to "transfer" the new command to the old command
                        add_to_last_cmd = False
                        last_cmd = va_cmd[-1]
                        if last_cmd.have_cmd(PattVecCmd.CMD_VECTOR) == False:
                            add_to_last_cmd = True
                                
                        proc_commands = self.proc2cmd[fpn]
                        last_start_loop_va = -1
                        m_va = 0
                        for proc_command in proc_commands:
                            if add_to_last_cmd:
                                cids = proc_command.get_cmd_ids()
                                for cid in cids: 
                                    value = proc_command.get_value(cid)
                                    last_cmd.add_cmd(cid, value)
                                add_to_last_cmd = False
                            else:
                                va_cmd.append(proc_command)

                                is_shift_start = False
                                is_shift_stop = False

                                for cmd in proc_command.get_cmd_ids():
                                    cmd_name = PattVecCmd.get_cmd_name(cmd)
                                    if cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_START_LOOP]:
                                        last_start_loop_va = m_va
                                    elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_START_SHIFT]:
                                        is_shift_start = True
                                    elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP_SHIFT]:
                                        is_shift_stop = True
                                    elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP_LOOP]:
                                        jmp = m_va - last_start_loop_va - 1
                                        va_cmd[-1].set_value(PattVecCmd.CMD_STOP_LOOP, jmp)
                                        last_start_loop_va = -1
                                    elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_START_MATCHLOOP]:
                                        last_start_loop_va = m_va
                                    elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP_COUNT_MATCHLOOP]:
                                        jmp = m_va - last_start_loop_va - 1
                                        va_cmd[-1].set_value(PattVecCmd.CMD_STOP_COUNT_MATCHLOOP, jmp)
                                        last_start_loop_va = -1
                                    elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_STOP_INF_MATCHLOOP]:
                                        jmp = m_va - last_start_loop_va - 1
                                        va_cmd[-1].set_value(PattVecCmd.CMD_STOP_INF_MATCHLOOP, jmp)
                                        last_start_loop_va = -1
                                    elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_GOTO]:
                                        va_cmd[-1].set_value(PattVecCmd.CMD_GOTO, proc_command.get_value(cmd))
                                    elif cmd_name == PattVecCmd.cmds[PattVecCmd.CMD_LABEL]:
                                        va_cmd[-1].set_value(PattVecCmd.CMD_LABEL, proc_command.get_value(cmd))

                                if is_shift_start and is_shift_stop:
                                    if is_wfc_subs:
                                        for i in range(new_vectors-1):
                                            cmd = PattVecCmd()
                                            cmd.add_cmd(PattVecCmd.CMD_VECTOR)
                                            va_cmd.append(cmd)
                                elif is_shift_start or is_shift_stop:
                                    int_err_msg = "INTERNAL COMPILER ERROR : Shift block with more of one vector is not yet supported!"
                                    raise STILDumpCompilerException(-1, -1, int_err_msg)


                                m_va += 1

                        for i in range(proc_va_length):
                            vap.append("-")
                            vamp.append(i)
                            
        
        self.data.append(va_cmd)
        
        if len(va_cmd) != exp_va:
            int_err_msg = f"INTERNAL COMPILER ERROR : total vector addresses ({exp_va}) does not match with number of commands ({len(va_cmd)})"
            raise STILDumpCompilerException(-1, -1, int_err_msg)
        if len(vap) != exp_va:
            int_err_msg = f"INTERNAL COMPILER ERROR : total vector addresses ({exp_va}) does not match pattern VA ({len(vap)})"
            raise STILDumpCompilerException(-1, -1, int_err_msg)
        if len(vamp) != exp_va:
            int_err_msg = f"INTERNAL COMPILER ERROR : total vector addresses ({exp_va}) does not match proc/macro VA ({len(vamp)})"
            raise STILDumpCompilerException(-1, -1, int_err_msg)
        
            
        # If there are commands which expands VA like Macro,
        # we have to update the vector address columns as well
        if exp_va > 0:
                        
            #print(f"data {data}")
            self.data[0] = []
            self.data[0] = range(exp_va)
    
            new_sep = ['|']*(exp_va)
            self.data[1] = []
            self.data[1] = new_sep
    
            self.data[2] = []
            self.data[2] = vap
    
            self.data[3] = []
            self.data[3] = new_sep
    
            self.data[4] = []
            self.data[4] = vamp
    
            self.data[5] = []
            self.data[5] = new_sep
    
            self.data[6+len(signals)] = []
            self.data[6+len(signals)] = new_sep
            
        if self.dump_data:
            self.dump_pattern_data(file_name, 
                                   signals, 
                                   block_name, 
                                   timing, 
                                   sig_group_domain, 
                                   macro_domain, 
                                   proc_domain, 
                                   exp_va, 
                                   block_scan_mem)
        elif is_pattern_block:
            # "compile mode", implement your code to compile self.data into 
            # specific FW/FPGA commands and data
            self.compile_data(  signals, 
                                timing, 
                                block_scan_mem)

        va_size = len(list(self.data[0]))

        if is_pattern_block:
            return va_size, proc_va
        else:
            return va_size

    def compile_data(self, signals, timing, block_scan_mem):
        '''
        Empty function for compilation self.data into your specific FW/FPGA commands and data

        Parameters
        ----------
        signals : list
            List of signal names.
        timing : string
            Time domain name.
        block_scan_mem : map
            Key is the signal name
            Value is a list with scan chain memory filled with WFC.

        Returns
        -------
        None.

        
        self.data contains list of lists with :
        [0]     VA - list with vector addresses
        [1]     | separator 
        [2]     VAP - list with pattern only vector addresses
        [3]     | separator 
        [4]     VAMP -list with macro/proc only vector addresses
        [5]     | separator 
        [6+X-1]   For every used signal (total X signals), separate list with own WFCs
        [6+X]   | separator 
        [6+X+1] PattVecCmd for every vector.
        '''
        pass

    def dump_pattern_data(self, file_name, signals, block_name, timing, sig_group_domain, macro_domain, proc_domain, exp_va, block_scan_mem):
        
        file = os.path.join(self.out_folder, file_name)
        f = open(file, "w")

        # start the dump:
        # Write first description:
        f.write(f"# Block name  : {block_name}\n\n")
        f.write( "# Domain info:\n")
        f.write(f"# - Timing domain       : {timing}\n")
        f.write(f"# - SignalsGroup domain : {sig_group_domain}\n")
        f.write(f"# - MacroDefs domain    : {macro_domain}\n")
        f.write(f"# - Procedures domain   : {proc_domain}\n\n")

        f.write("# Format of compiled data :\n")
        f.write("# VA|VAP|VAMP|WFC|CMD|CMD arg| .... |CMD|CMD arg \n\n")

        
        f.write( "# Vector data information :\n")
        f.write( "# VA   - absolute vector address for this particular pattern block. \n\
        This address can be used also as relative vector address when all \n\
        pattern blocks are assembled into the memory.\n")
        f.write( "# VAP  - absolute vector address in the pattern block if there is no \n\
        Macro or expanding Procedure invocation in this pattern block.\n\
        Value '-' indicates that there is WFC/CMD data from Macro/Procdure at this VA\n")
        f.write( "# VAMP - absolute vector address of the Macro or expanded Procedure block.\n\
        Value '-' indicates that there is no WFC/CMD data from Macro/Procdure at this VA\n\n")

        f.write( "# Commands syntax:\n")
        f.write( "# COMMAND_NAME=value_if_applicable\n")
        f.write( "# If more than one command exists on the same vector address, they are separated with | \n\n")

        f.write("# SIGNALS_ORDER|")
        s = ""
        si = 0
        err_msg = ""
        raise_int_err = False
        for signal in signals:
            s += signal + "+"
            wfc_len = len(self.data[si+6])
            if wfc_len != exp_va:
                err_msg += f"INTERNAL COMPILER ERROR : total vector addresses ({exp_va}) does not match WFC length ({wfc_len}) of the signal {signal}\n"
                raise_int_err = True
            si += 1
        f.write(f"{s[:-1]}")
        f.write("|\n\n")
        
        if raise_int_err:
            raise STILDumpCompilerException(-1, -1, err_msg)

        """
         Finally dump all data like in text version of "memory map"
         Transponse VA, WFC and commands data from columns (lists) to rows: 
         VA ->         [  0,     1,     2  ...,  N]
         WFC -> sig1 = [ "L",   "H",   "H" ..., "X"]
         WFC -> sig2 = [ "0",   "1",   "1" ..., "Z"]
         CMD -> cmd  = [ "W" , "SL|5",  "" ..., "L|label1" ]
         after transponse:
         
         0|L0|W
         1|H1|SL|5
         2|H1|
         .....
         N|XZ|L|label1
         
         For large STIL file, avoid using numpy.transpose() method. 
         It use a huge amount of RAM and in result is very slow:
         mem_map = list(map(list, np.transpose(data)))
         for va in mem_map:
             for mem_data in va:
                 f.write(f"{mem_data}")
             f.write("\n")         
        """        
        f.write("# VA| VAP|VAMP|  WFCs  |CMD ... \n")
        va_size = len(list(self.data[0]))
        for va in range(va_size):
            s =""
            c = 0
            for column in self.data:
                if len(column) > 0:
                    if c == 0 or c == 2 or c == 4:
                        s += str(column[va]).rjust(4)
                    elif c == 6:
                        s += str(column[va]).rjust(6)
                    else:
                        s += str(column[va])
                c += 1
                if c > (6+si+1):
                    break
            s += "\n"
            f.write(s)
        f.close()

        if self.is_scan_mem_available:
            scan_mem_file = file_name + ".sm"
            file = os.path.join(self.out_folder, scan_mem_file)
            f = open(file, "w")

            # write header before the data
            f.write("# SIGNAL_NAME | WFC sequence for the scan memory\n")

            for sig in block_scan_mem:
                f.write(sig)
                f.write("|")
                wfcs = block_scan_mem[sig]
                for wfc in wfcs:
                    f.write(wfc)
                f.write("\n")
                    
            f.close()
    
    def collect_wfc_for_signals(self, s2w, sgds, block_name, sig_order, wfc_order):

        # For each signal group domain used by macro/proc/pattern block
        for sgd in sgds:

            last_sig2wfc = None
            if sgd in self.sgd2last_sig2wfc:
                last_sig2wfc = self.sgd2last_sig2wfc[sgd]

            sig2wfc_before_subs = None
            if sgd not in self.sgd2before_subs_sig2wfc:
                self.sgd2before_subs_sig2wfc[sgd] = {}
            sig2wfc_before_subs = self.sgd2before_subs_sig2wfc[sgd]

            sig2wfc = WFCUtils.collect_sig2wfc(self.sig2type.keys(),
                                               sgd,
                                               self.signal_groups2signals,
                                               sig_order,
                                               wfc_order,
                                               last_sig2wfc,
                                               sig2wfc_before_subs)

            self.sgd2last_sig2wfc[sgd] = sig2wfc

            sgd_block = sgd + "::" + block_name
            if sgd_block not in s2w:
                s2w[sgd_block] = {}
            
            if self.is_first_vector and self.is_vector_stmt:
                s2w[sgd_block] = sig2wfc
            else:
                old_sig2wfc = s2w[sgd_block]
                for sig in sig2wfc:
                    wfc = sig2wfc[sig]
                    if sig not in old_sig2wfc:
                        old_sig2wfc[sig] = wfc
                    else:
                        old_sig2wfc[sig] += wfc

    
    def b_macrodefs__pattern_statements__CLOSE_VECTOR_BLOCK(self, t):

        if self.is_vector_stmt:
            self.save_cmd_macro()

        sgds = self.macro2signal_group_domains[self.curr_macro_name]
        
        self.collect_wfc_for_signals(self.sgd_macro2sig2WFCs,
                                     sgds,
                                     self.curr_macro_name,
                                     self.curr_macro_sig_order,
                                     self.curr_macro_wfc_order
                                     )

        super().b_macrodefs__pattern_statements__CLOSE_VECTOR_BLOCK(t)

    def b_procedures__pattern_statements__CLOSE_VECTOR_BLOCK(self, t):

        if self.is_vector_stmt:
            self.save_cmd_proc()

        sgds = self.proc2signal_group_domains[self.curr_proc_name]

        self.collect_wfc_for_signals(self.sgd_proc2sig2WFCs,
                                     sgds,
                                     self.curr_proc_name,
                                     self.curr_proc_sig_order,
                                     self.curr_proc_wfc_order
                                     )

        
        super().b_procedures__pattern_statements__CLOSE_VECTOR_BLOCK(t)

    def dump_timing_data(self):
        
        if self.dump_data:
            timing_file = os.path.join(self.out_folder, "timing.txt")
            with open(timing_file, "w") as f:
                f.write("# Format of timing data :\n")
                f.write("# Timing domain name | WFT name | Signal name | WFC | WFE : Time .... \n")
                for sig_wft in self.sig_wft2timing:
                    data = sig_wft.split("::")
                    t_domain = DomainUtils.get_domain(data[0], True)

                    sti = self.sig_wft2timing[sig_wft]
                    wfcs = sti.get_wfcs()
                    for wfc in wfcs:
                        timing = sti.get_timing_for_wfc(wfc)
                        f.write(f"{t_domain} | {data[1]} | {data[2]} | {wfc} | ")
                        for wfe_time in timing:
                            wfe = wfe_time[0]
                            time = wfe_time[1]
                            f.write(f"{wfe}:{time} | ")
                        f.write("\n")

    def assamble_all(self):
        # not implemented yet
        pass

    def calculate_test_cycles(self):

        if self.is_parsing_done == False:
            return

        # Number of test cycles
        # test cycles != vector addresses!!!
        tc = 0

        for pattern_exec_block in self.patt_exec_block2patt_burst.keys():

            patt_bursts = self.patt_exec_block2patt_burst[pattern_exec_block]
            # For every pattern burst block in the pattern exec block
            for patt_burst in patt_bursts:

                pattern_list = self.patt_burst_block2pattern_blocks[patt_burst]
                proc_domain = self.patt_burst2proc_domain[patt_burst]
                macro_domain = self.patt_burst2macro_domain[patt_burst]
                sig_group_domain = self.patt_burst2sig_groups_domain[patt_burst]

                # For every pattern block in the pattern burst block
                for patt in pattern_list:

                    start_loop_va = 0
                    loop_count = 0

                    cmds_list = self.patt2cmd[patt]

                    for vec_cmds in cmds_list:
                        for cmd in vec_cmds.get_cmd_ids():
                            pm_name = ""
                            if (cmd == PattVecCmd.CMD_CALL) or (cmd == PattVecCmd.CMD_MACRO):
                                pm_name = vec_cmds.get_value(cmd)
                            props = vec_cmds.get_props(cmd)
                            sig2wfc_count = {}
                            if props is not None:
                                for key, value in props.items():
                                    if (
                                        cmd == PattVecCmd.CMD_CALL
                                        or cmd == PattVecCmd.CMD_MACRO
                                    ):
                                        sig2wfc_count[key] = len(value)
                            if cmd == PattVecCmd.CMD_VECTOR:
                                tc += 1
                            elif (
                                cmd == PattVecCmd.CMD_START_LOOP
                                or cmd == PattVecCmd.CMD_START_MATCHLOOP
                            ):
                                start_loop_va = tc
                                if vec_cmds.get_value(cmd) == "Infinite":
                                    loop_count = -1
                            elif (
                                cmd == PattVecCmd.CMD_LOAD_LOOP_COUNTER
                                or cmd == PattVecCmd.CMD_LOAD_MATCHLOOP_COUNTER
                            ):
                                loop_count = int(vec_cmds.get_value(cmd))
                            elif (
                                cmd == PattVecCmd.CMD_STOP_LOOP
                                or cmd == PattVecCmd.CMD_STOP_COUNT_MATCHLOOP
                            ):
                                if loop_count > 0:
                                    tc += (tc - start_loop_va + 1) * (loop_count - 1)
                                start_loop_va = 0

                            elif cmd == PattVecCmd.CMD_CALL:

                                proc_name = pm_name

                                pfn = DomainUtils.get_full_name(proc_domain, proc_name)

                                tc += self.proc2tc[pfn]

                                if pfn in self.proc_with_shift:

                                    max_va = 0

                                    for sig in sig2wfc_count:

                                        wfc_c = sig2wfc_count[sig]
                                        his = self.hashinfo[pfn]
                                        hash_not_in_shift = 0

                                        for hi in his:
                                            if hi.sig_ref in self.sig2type:
                                                if hi.pos_rel_shift != 1:
                                                    hash_not_in_shift += 1
                                            else:
                                                sg = (
                                                    sig_group_domain + "::" + hi.sig_ref
                                                )
                                                sig_gr = self.signal_groups2signals[sg]
                                                if sig in sig_gr:
                                                    if hi.pos_rel_shift != 1:
                                                        hash_not_in_shift += 1
                                        # number of WFC argument - # not in shift - shift Vector
                                        va_shift = wfc_c - hash_not_in_shift - 1
                                        if max_va == 0:
                                            max_va = va_shift
                                        elif va_shift > max_va:
                                            max_va = va_shift
                                    tc += max_va

                            elif cmd == PattVecCmd.CMD_MACRO:

                                macro_name = pm_name
                                mfn = DomainUtils.get_full_name(
                                    macro_domain, macro_name
                                )
                                tc += self.macro2tc[mfn]
                                if mfn in self.macro_with_shift:
                                    max_va = 0
                                    for sig in sig2wfc_count:

                                        wfc_c = sig2wfc_count[sig]
                                        his = self.hashinfo[mfn]
                                        hash_not_in_shift = 0

                                        for hi in his:
                                            if hi.sig_ref in self.sig2type:
                                                pass
                                            else:
                                                sg = (
                                                    sig_group_domain + "::" + hi.sig_ref
                                                )
                                                sig_gr = self.signal_groups2signals[sg]
                                                if sig in sig_gr:
                                                    if hi.pos_rel_shift != 1:
                                                        hash_not_in_shift += 1
                                        # number of WFC argument - # not in shift - shift Vector
                                        va_shift = wfc_c - hash_not_in_shift - 1
                                        if max_va == 0:
                                            max_va = va_shift
                                        elif va_shift > max_va:
                                            max_va = va_shift
                                    tc += max_va

        print(f"\nTest cycles = {tc}\n")



# -*- coding: utf-8 -*-
import inspect

from .DomainUtils import DomainUtils
from .WFCUtils import WFCUtils
from .SigTimingInfo import SigTimingInfo


class TimingBlockParser:

    force_events = [
        "ForceDown",
        "D",
        "ForceUp",
        "U",
        "ForceOff",
        "Z",
        "ForcePrior",
        "P",
    ]
    compare_events = [
        "CompareLow",
        "L",
        "CompareHigh",
        "H",
        "CompareUnknown",
        "X",
        "x",
        "CompareOff",
        "T",
        "CompareValid",
        "V",
        "CompareLowWindow",
        "l",
        "CompareOff",
        "h",
        "CompareHighWindow",
        "h",
        "CompareOffWindow",
        "t",
        "CompareValidWindow",
        "v",
        
    ]

    def __init__(self, debug=False):
        self.debug = debug
        # List with all parsed timing domains
        self.timing_domains = []

        self.curr_signal_groups_domain = DomainUtils.global_domain
        self.curr_timing_domain = DomainUtils.global_domain

        # signal expression references in Timing block
        # can be either single signal or group of signals
        self.curr_sigref = []

        # dict key   is the signal in full domain name : timing_domain::signal_name
        # dict value is a list with all WFC defined for this signal/domain
        self.sig2wfc = {}

        # dict key   is the timing_domain::wft::signal_name
        # dict value is a SigTimingInfo object with timing information 
        self.sig_wft2timing = {}

        # dict key   is the timing domain name
        # dict value is a list with all waveform tables (WFT)
        self.time_domain2wft = {}

        self.curr_wft = ""

        self.curr_wfc_list = ""

        self.curr_wfe_count = 0

        self.curr_sig_type = ""

        self.curr_events = []

        # dict key   is the timing_domain::wft
        # dict value is the wft's period value.
        self.wft2period = {}

        self.is_timing_block_defined = False

    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def b_timing__TIMING_DOMAIN_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_timing_domain = t.value

    def b_timing__OPEN_TIMING_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        if self.is_signal_block_defined == False:
            err_msg = "Expected Signal block before any Timing block"
            raise Exception(err_msg)
        if self.is_signalgroups_block_defined == False:
            err_msg = "Expected SignalGroups block before any Timing block"
            raise Exception(err_msg)
        self.is_timing_block_defined = True

        if self.curr_timing_domain in self.timing_domains:
            domain_name = DomainUtils.get_domain(self.curr_timing_domain, True)
            err_msg = f"Timing block domain '{domain_name}' is already defined !"
            raise Exception(err_msg)
        else:
            self.timing_domains.append(self.curr_timing_domain)

    def b_timing__WFT_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_wft = t.value

        if self.curr_timing_domain in self.time_domain2wft:
            wft_list = self.time_domain2wft[self.curr_timing_domain]
            if self.curr_wft in wft_list:
                err_msg = f"Waveform Table named '{self.curr_wft}' is already defined"
                raise Exception(err_msg)
            else:
                wft_list.append(self.curr_wft)
        else:
            self.time_domain2wft[self.curr_timing_domain] = [self.curr_wft]

    def b_timing__period(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
            
        key = self.curr_timing_domain + "::" + self.curr_wft
        self.wft2period[key] = t[1].value
        
    def b_timing__SIGNAL_GROUPS_DOMAIN(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        sig_groups = t.value
        if sig_groups not in self.signal_groups_domains:
            err_msg = f"Signal groups domain {sig_groups} is not defined"
            raise Exception(err_msg)
        self.curr_signal_groups_domain = t.value

    def b_timing__WF_SIGREF_EXPR(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        sig_ref = t.value

        if sig_ref in self.sig2type:
            self.curr_sig_type = self.sig2type[sig_ref]
            ftn = DomainUtils.get_full_name(self.curr_timing_domain, sig_ref)
            self.curr_sigref.append(ftn)
        else:
            sig_group_domain = self.curr_signal_groups_domain
            fsn = DomainUtils.get_full_name(sig_group_domain, sig_ref)
            if fsn in self.signal_groups2signals:
                signals = self.signal_groups2signals[fsn]
                sig_type = ""
                for signal in signals:

                    if sig_type != "":
                        new_sig_type = self.sig2type[signal]
                        if sig_type == "In" and new_sig_type == "Out":
                            err_msg = "Signal '{sig_ref}' contain both input and output signals, which can not be used for defining WaveForm Characters!"
                            raise Exception(err_msg)
                        elif sig_type == "Out" and new_sig_type == "In":
                            err_msg = "Signal '{sig_ref}' contain both input and output signals, which can not be used for defining WaveForm Characters!"
                            raise Exception(err_msg)
                        pass
                    else:
                        st = self.sig2type[signal]
                        if st == "In" or st == "Out":
                            sig_type = st

                    ftn = DomainUtils.get_full_name(self.curr_timing_domain, signal)
                    self.curr_sigref.append(ftn)
            else:
                domain = DomainUtils.get_domain(sig_group_domain, True)
                err_msg = f"Signal {sig_ref} from '{domain}' domain is not defined"
                raise Exception(err_msg)

    def check_if_wfc_defined(self, sigref, wfc):

        sig = DomainUtils.get_name(sigref)
        full_sig_name = self.curr_timing_domain + '::' + self.curr_wft + '::' + sig
        
        if full_sig_name in self.sig_wft2timing:
            sig_time_info = self.sig_wft2timing[full_sig_name]
            if sig_time_info is not None:
                timing = sig_time_info.get_timing_for_wfc(wfc)
                if timing is not None:
                    err_msg = f"WFC {wfc} is already defined for signal {sig}"
                    raise Exception(err_msg)
            

    def b_timing__WFC_LIST(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.curr_wfc_list = t.value

        for sigref in self.curr_sigref:
            wfcs = WFCUtils.trim_wfcs(self.curr_wfc_list)
            if sigref in self.sig2wfc:
                for wfc in wfcs:

                    self.check_if_wfc_defined(sigref, wfc)
                    
                    stored_wfc = self.sig2wfc[sigref]
                    if wfc not in stored_wfc:
                        self.sig2wfc[sigref] += wfc
            else:
                self.sig2wfc[sigref] = wfcs

    def b_timing__EVENT(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
            
        wfe = t.value

        if wfe != "/":
            self.curr_wfe_count += 1
            if self.curr_sig_type == "In":
                if wfe not in TimingBlockParser.force_events:
                    err_msg = f"Compare Waveform event '{wfe}' is used for 'In' type of signal !"
                    raise Exception(err_msg)
            elif self.curr_sig_type == "Out":
                if wfe not in TimingBlockParser.compare_events:
                    err_msg = f"Force Waveform event '{wfe}' is used for 'Out' type of signal !"
                    raise Exception(err_msg)
        self.curr_events.append(wfe)

    def b_timing__time_offset(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        wfe_count = self.curr_wfe_count
        wfc_count = len(self.curr_wfc_list)
        if wfe_count > 1 and wfe_count != wfc_count:
            err_msg = f"Number of Waveform Events ({wfe_count}) does not match with number of Waveform Chars ({wfc_count})"
            raise Exception(err_msg)
            
        for domain_sig in self.curr_sigref:
            
            sig = DomainUtils.get_name(domain_sig)
            
            full_sig_name = self.curr_timing_domain + '::' + self.curr_wft + '::' + sig
            
            if full_sig_name not in self.sig_wft2timing:
                sig_time_info = SigTimingInfo(sig)
                self.sig_wft2timing[full_sig_name] = sig_time_info
            else:
                sig_time_info = self.sig_wft2timing[full_sig_name]
            
            index = 0
            for wfc in self.curr_wfc_list:
                if len(self.curr_events) == 1:
                    sig_time_info.set_timing_for_wfc(wfc, self.curr_events[0], t[0] )
                else:
                    sig_time_info.set_timing_for_wfc(wfc, self.curr_events[index], t[0] )
                index += 1
            

        self.curr_events.clear()
        self.curr_wfe_count = 0
        self.curr_sig_type = ""

    def b_timing__CLOSE_WFC_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
            
        self.curr_sigref.clear()
        self.curr_wfc_list = ""
        self.curr_events.clear()

    def b_timing__CLOSE_TIMING_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_timing_domain = DomainUtils.global_domain
        self.curr_wft = ""



# -*- coding: utf-8 -*-
import io

from setuptools import find_packages, setup
from Semi_ATE.STIL import __version__

# =============================================================================
# Use README.md for long description
# =============================================================================
with io.open("README.md", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

# https://setuptools.readthedocs.io/en/latest/
# PEP-314 (Metadata for Python Software Packages) : https://www.python.org/dev/peps/pep-0314/
# https://pypi.org/classifiers/

setup(
    name="Semi-ATE-STIL",
    version=__version__,
    description="Standard Tester Interface Library [IEEE1450] with python",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Tom Hören",
    maintainer="Semi-ATE",
    maintainer_email="info@Semi-ATE.com",
    url="https://github.com/Semi-ATE/STIL",
    license="GPLv2",
    keywords="Semiconductor ATE Automatic Test Equipment STIL IEEE1450",
    platforms=["Windows", "Linux", "Mac OS-X"],
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Manufacturing",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)",
        "Topic :: Software Development :: Code Generators",
        "Topic :: Software Development :: Compilers",
        "Topic :: Software Development :: Debuggers",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# -*- coding: utf-8 -*-
class STILSemanticException(Exception):
    def __init__(self, err_line, err_col, err_msg):
        print(f"new STILSemanticException {err_line} {err_col} {err_msg}")
        self.line = err_line 
        self.col  = err_col 
        self.msg  = err_msg 



# -*- coding: utf-8 -*-

#    Work in process ...


error_map = {
    "base__KEYWORD_ANN": "Ann",
    "b_stil__KEYWORD_STIL": "STIL",
    "b_stil__STIL_VERSION": "version of the STIL = 1.0",
    "b_stil__STIL_EXTENSION_NAME": "DCLevels or Design",
    "b_stil__STIL_EXTENSION_YEAR": "2002 or 2005",
    "b_header__KEYWORD_HEADER": "Header",
    "b_header__KEYWORD_TITLE": "Title",
    "b_header__KEYWORD_SOURCE": "Source",
    "b_header__KEYWORD_DATE": "Date",
    "b_header__KEYWORD_HISTORY": "History",
    "b_header__HEADER_DATE_STRING": "Data/Time information in double quoted string",
    "b_header__TITLE_STRING":  "Title information in double quoted string",
    "b_header__SOURCE_STRING": "Source information in double quoted string",
    "b_signals__SIGNAL_TYPE": "signal type : In, Out, InOut, Pseudo or Supply",
    "b_timing__WFT_NAME": "Name of the WaveformTable",
    "b_pattern_burst__PATTERN_BURST_BLOCK_NAME" : "Name of the PatternBurst block",
    "b_pattern_exec__PATTERN_EXEC_BLOCK_NAME" : "Name of the PatternExec block",
    "b_macrodefs__MACRO_NAME" : "Name of the MacroDefs block",
    "b_pattern__pattern_statements__USER_KEYWORD_VALUE" : "UserKeyword value",
    "b_procedures__pattern_statements__USER_KEYWORD_VALUE" : "UserKeyword value",
    "b_macrodefs__pattern_statements__USER_KEYWORD_VALUE" : "UserKeyword value",
    "b_pattern__pattern_statements__UK_BLOCK_VALUE" : "value in the UserKeyword block",
    "b_procedures__pattern_statements__UK_BLOCK_VALUE" : "value in the UserKeyword block",
    "b_macrodefs__pattern_statements__UK_BLOCK_VALUE" : "value in the UserKeyword block",
    "ESCAPED_STRING" : "String in double quotes"    
}


class SyntaxParserExceptions:
    """
    Class for transforming automatically generated
    Lark syntax errors in human readable text
    Work in process ...
    """

    def __init__(self, debug=False):
        self.debug = debug

    def transform(self, lark_error):

        if self.debug:
            print(f"Catched syntax error(s) from lark parser:\n{lark_error}")

        if len(lark_error) == 1:
            return_value = "Expected : "
        elif len(lark_error) > 1:
            return_value = "Expected one of : "
        else:
            print(lark_error)
#            raise Exception()
            return "Lark parser does not provide expected value, but should!"

        is_open_block_used = False
        for err in lark_error:
            if self.debug:
                print(f"Error {err}")
            v = error_map.get(err)
            # If the error can not be found in the error map:
            if v == None:
                if err.endswith("USER_DEFINED_NAME"):
                    return_value += "User defined name"
                elif err.endswith("TIME_EXPR"):
                    return_value += "Time expression in format 'NUMBER UNIT' without whitespace between them or 'NUMBER MATH_OPERATOR NUMBER UNIT' or 'NUMBER' or 'VARIABLE' \n\t\t   Where:\n\t\t   - Mandatory single quotes 'xxx' for the value\n\t\t   - NUMBER can be integer, real or exponential number\n\t\t   - UNIT SI Time/Frequency unit with engineering prefixes (ns, us, ms, s, Hz, MHz ...)\n\t\t   - VARIABLE defined in the Spec block\n\t\t   - MATH_OPERATOR can be *,/,+.- \n\t\t   Examples: '5ns' '1.234ms' '10E-3s' '1MHz' '2/1MHz+PLL_Freq' "
                elif err.endswith("SEMICOLON"):
                    return_value += ";"
                elif err.find("QUOTE") > -1:
                    return_value += "'"
                elif err.find("EQUAL") > -1:
                    return_value += "="
                elif err.find("COLON") > -1:
                    return_value += ":"
                elif err.find("SLASH") > -1:
                    return_value += "/"
                elif err.find("TIMEUNIT") > -1:
                    return_value += "TimeUnit"
                elif err.endswith("_ADD"):
                    return_value += "+"
                elif err.endswith("_SUB"):
                    return_value += "-"
                elif err.find("__OPEN_") > -1:
                    if is_open_block_used == False:
                        return_value += "{"
                        is_open_block_used = True
                    else:
                        continue
                elif err.find("__CLOSE_") > -1:
                    return_value += "}"
                elif err.find("__KEYWORD_ANN") > -1:
                    return_value += "Ann"
                elif err.endswith("__INT"):
                    return_value += "integer value"
                elif err.endswith("_DOMAIN_NAME"):
                    return_value += "domain name"
                elif err.endswith("_END_STMT"):
                    return_value += ";"
                elif err.endswith("SIGREF_NAME"):
                    return_value += "signal name"
                elif err.endswith("KEYWORD_BLOCK_SIGNAL_GROUPS"):
                    return_value += "SignalGroups"
                elif err.endswith("SIGNAL_GROUPS_DOMAIN"):
                    return_value += "Name of the SignalGroups domain"                  
                elif err.endswith("KEYWORD_BASE"):
                    return_value += "Base"
                elif err.endswith("KEYWORD_DATABITCOUNT"):
                    return_value += "DataBitCount"
                elif err.endswith("KEYWORD_SCANOUT"):
                    return_value += "ScanOut"
                elif err.endswith("KEYWORD_SCANIN"):
                    return_value += "ScanIn"
                elif err.endswith("KEYWORD_TERMINATION"):
                    return_value += "Termination"
                elif err.endswith("KEYWORD_DEFAULT_STATE"):
                    return_value += "DefaultState"
                elif err.endswith("KEYWORD_ALIGNMENT"):
                    return_value += "Alignment"
                elif err.endswith("KEYWORD_START"):
                    return_value += "Start"
                elif err.endswith("KEYWORD_STOP"):
                    return_value += "Stop"
                elif err.endswith("KEYWORD_PROCEDURES"):
                    return_value += "Procedures"
                elif err.endswith("PROCEDURES_DOMAIN"):
                    return_value += "Name of the Procedures domain"
                elif err.endswith("KEYWORD_MACRO_DEFS"):
                    return_value += "MacroDefs"
                elif err.endswith("MACROS_DOMAIN"):
                    return_value += "Name of the MacroDefs domain"
                elif err.endswith("KEYWORD_SIGNAL_GROUPS"):
                    return_value += "SignalGroups"
                elif err.endswith("KEYWORD_IFNEEDED"):
                    return_value += "IfNeed"
                elif err.endswith("KEYWORD_BLOCK_SCAN_STRUCTURES"):
                    return_value += "ScanStructures"
                elif err.endswith("KEYWORD_BLOCK_ANN"):
                    return_value += "Ann"
                elif err.endswith("KEYWORD_BLOCK_USER_KEYWORDS"):
                    return_value += "UserKeywords"
                elif err.endswith("KEYWORD_BLOCK_USER_FUNCTIONS"):
                    return_value += "UserFunctions"
                elif err.endswith("KEYWORD_BLOCK_PATTERN"):
                    return_value += "Pattern"
                elif err.endswith("KEYWORD_BLOCK_SELECTOR"):
                    return_value += "Selector"
                elif err.endswith("KEYWORD_BLOCK_PATERN_EXEC"):
                    return_value += "PatternExec"
                elif err.endswith("KEYWORD_BLOCK_SPEC"):
                    return_value += "Spec"
                elif err.endswith("KEYWORD_BLOCK_SIGNALS"):
                    return_value += "Signals"
                elif err.endswith("KEYWORD_BLOCK_HEADER"):
                    return_value += "Header"
                elif err.endswith("KEYWORD_BLOCK_PROCEDURES"):
                    return_value += "Procedures"
                elif err.endswith("KEYWORD_BLOCK_TIMING"):
                    return_value += "Timing"
                elif err.endswith("KEYWORD_BLOCK_PATTERN_BURST"):
                    return_value += "PatternBurst"
                elif err.endswith("KEYWORD_BLOCK_MACRO_DEFS"):
                    return_value += "MacroDefs"
                elif err.endswith("KEYWORD_PAT_LIST"):
                    return_value += "PatList"
                elif err.endswith("KEYWORD_CALL"):
                    return_value += "Call"
                elif err.endswith("KEYWORD_BREAKPOINT"):
                    return_value += "BreakPoint"
                elif err.endswith("KEYWORD_CONDITION"):
                    return_value += "Condition"
                elif err.endswith("KEYWORD_C"):
                    return_value += "C"
                elif err.endswith("USER_KEYWORD"):
                    return_value += "UserKeyword"
                elif err.endswith("KEYWORD_IDDQ_TEST_POINT"):
                    return_value += "IddqTestPoint"
                elif err.endswith("KEYWORD_MATCH_LOOP"):
                    return_value += "MatchLoop"
                elif err.endswith("KEYWORD_LOOP"):
                    return_value += "Loop"
                elif err.endswith("KEYWORD_W"):
                    return_value += "W"
                elif err.endswith("KEYWORD_WAVEFORM_TABLE"):
                    return_value += "WaveformTable"
                elif err.endswith("KEYWORD_MACRO"):
                    return_value += "Macro"
                elif err.endswith("LABEL"):
                    return_value += "label name"
                elif err.endswith("KEYWORD_GOTO"):
                    return_value += "Goto"
                elif err.endswith("KEYWORD_FIXED"):
                    return_value += "Fixed"
                elif err.endswith("KEYWORD_F"):
                    return_value += "F"
                elif err.endswith("KEYWORD_VECTOR"):
                    return_value += "Vector"
                elif err.endswith("KEYWORD_V"):
                    return_value += "V"
                elif err.endswith("KEYWORD_SHIFT"):
                    return_value += "Shift"
                elif err.endswith("KEYWORD_SCAN_CHAIN"):
                    return_value += "ScanChain"
                elif err.endswith("KEYWORD_SCAN_CHAIN"):
                    return_value += "ScanChain"
                elif err.endswith("KEYWORD_SCAN_STRUCTURES"):
                    return_value += "ScanStructures"
                elif err.endswith("ACTIVESCANCHAINS"):
                    return_value += "ActiveScanChains"
                elif err.endswith("SCAN_STRUCTURES_DOMAIN"):
                    return_value += "Name of the ScanStructures domain"
                elif err.endswith("TERMINATEHIGH"):
                    return_value += "TerminateHigh"
                elif err.endswith("TERMINATELOW"):
                    return_value += "TerminateLow"
                elif err.endswith("TERMINATEOFF"):
                    return_value += "TerminateOff"
                elif err.endswith("TERMINATEUNKNOWN"):
                    return_value += "TerminateUnknown"
                elif err.endswith("SIG_DEF_STATE_UP"):
                    return_value += "U or ForceUp"
                elif err.endswith("SIG_DEF_STATE_OFF"):
                    return_value += "Z or ForceOff"
                elif err.endswith("SIG_DEF_STATE_DOWN"):
                    return_value += "D or ForceDown"
                elif err.endswith("HEX"):
                    return_value += "Hex"
                elif err.endswith("DEC"):
                    return_value += "Dec"
                elif err.endswith("WFC_LIST"):
                    return_value += "List of WFC"
                elif err.endswith("KEYWORD_PERIOD"):
                    return_value += "Period"
                elif err.endswith("SIGREF_EXPR"):
                    return_value += "Signal"
                elif err.endswith("EVENT"):
                    return_value += "one or more (separated by '/') Waveform events :\n\t\t   D,U,Z,P for drive \n\t\t   L,H,X,T,V,l,h,x,t,v for compare"
                elif err.endswith("LABEL_NAME"):
                    return_value += "Name of label"
                elif err.endswith("VEC_DATA_STRING"):
                    return_value += "List of WFC"
                elif err.endswith("WAVEFORM_TABLE_NAME"):
                    return_value += "Name of the WaveformTable"
                elif err.endswith("CALL_PROC_NAME"):
                    return_value += "Name of the Procedure"
                elif err.endswith("CALL_MACRO_NAME"):
                    return_value += "Name of the Macro"
                elif err.endswith("MATCHLOOP_INF"):
                    return_value += "Infinite"
                elif err.endswith("SCANCHAIN_NAME"):
                    return_value += "Name of the ScanChain"
                elif err.endswith("PATTERN_NAME"):
                    return_value += "Name of the Pattern block"
                elif err.endswith("LOOP_COUNT"):
                    return_value += "Loop count value"
                else:
                    return_value += err
            else:
                return_value += v
            return_value += " or "

        return_value = return_value[:-4]

        return return_value



#!/usr/bin/python
# -*- coding: utf-8 -*-
import os
import re

from lark import Lark
from lark.exceptions import UnexpectedToken
from lark.exceptions import UnexpectedCharacters
from lark.exceptions import VisitError

from .STILLark import STILLark

from . import SyntaxParserExceptions
from . import STILSemanticException
from . import utils

class STILParser(STILLark):
    def __init__(self, 
                 stil_file, 
                 propagate_positions=True, 
                 expanding_procs = False, 
                 stil_lark_file = None,
                 extra_grammars = [],
                 debug=False):

        STILLark.__init__(self, stil_file, debug)

        self.err_line = -1
        self.err_col = -1
        self.err_msg = ""

        self.tree = None

        self.is_parsing_done = False
        # Depends on ATE pattern engine capabilities, procedures can be expanded or not.
        # This choice impacts calculation of the total VA
        self.expanding_procs = expanding_procs

        grammar_base_file = os.path.dirname(__file__)
        # Adding main lark grammar file
        if stil_lark_file == None:
            grammar_file = os.path.join(str(grammar_base_file), "grammars", "stil.lark")
        else:
            grammar_file = stil_lark_file
                
        grammars_file = os.path.join(str(grammar_base_file), "grammars")
        # Adding all lark grammars files under grammars folder
        ip = list()
        ip.append(grammars_file)

        # Adding additional grammar files
        if extra_grammars:
            ip.extend(extra_grammars)
            
        with open(grammar_file) as grammar:
            self.parser = Lark(
                grammar.read(),
                start="start",
                parser="lalr",
                propagate_positions=propagate_positions,
                import_paths=ip,
            )


        self.is_compressed = utils.check_for_compression(stil_file)
    
    def preprocess_include(self, file):
        
        out_file = file + ".wo_include"

        f = open(out_file, 'wb')
        if f == None:
            print(f"ERROR: Can not create output file {out_file}.\n")
            return "ERROR"
        else:
            f.close()
            os.remove(out_file)
        
        with open(file) as data, open(out_file, "w") as out:
            inc_cnt = 0
            line_cnt = 0
            line = data.readline()
            out.write(line)
            err_inc_msg = "Not correct 'Include' syntax at line "
            err_file_not_found_msg = "Can not find include flie "
            while line:
                line = data.readline()
                line_cnt += 1
                strip_line = line.strip()
                if strip_line.startswith("Include"):
                    l = re.split(r'\s+', strip_line)
                    if len(l) > 1:
                        l = l[1].split('"')
                        inc_cnt += 1
                        if len(l) > 1:
                            inc_file = l[1]
                            folder = os.path.dirname(file)
                            f = os.path.join(folder, inc_file)
                            if os.path.exists(f):
                                with open(f) as inc_data:
                                    inc_line = inc_data.readline()
                                    sl = inc_line.strip()
                                    if sl.startswith("STIL") == False:
                                        out.write(inc_line)
                                    while inc_line:
                                        inc_line = inc_data.readline()
                                        sl = inc_line.strip()
                                        if sl.startswith("STIL") == False:
                                            out.write(inc_line)
                            else:
                                print(f"ERROR: {err_file_not_found_msg} '{inc_file}' at line {line_cnt}")
                                return "ERROR"
                        else:
                            print(f"ERROR: {err_inc_msg} {line_cnt}")
                            return "ERROR"
                    else:
                            print(f"ERROR: {err_inc_msg} {line_cnt}")
                            return "ERROR"
                else:
                    out.write(line)
        
        if inc_cnt > 0:
            return out_file
        else:
            os.remove(out_file)
            return None

    def parse_syntax(self, debug=False, preprocess_include = True):
        """
        Parameters
        ----------
        stil : str
            stil argument can be either location of STIL file or
            the STIL file content as string.
        debug : TYPE, optional
            Debug flag. The default is False.

        Returns
        -------
        tree : Lark tree object

        """
        if debug:
            print("\n===========================")
            print("Start of syntax parsing :\n")

        try:
            
            if os.path.exists(self.stil_file):
                if self.is_compressed:
                    data = utils.get_uncompressed_data(self.stil_file)
                    self.tree = self.parser.parse(data)
                    if debug == True:
                        print(self.tree.pretty())
                else:
                    
                    file = self.stil_file
                    
                    if preprocess_include:
                        # Initial support for 'Include'.
                        # Does not support syntax error highlighting yet.
                        full_file = self.preprocess_include(self.stil_file)
                        
                        if full_file != None:
                            if full_file == "ERROR":
                                return
                            else:
                                file = full_file
                        else:
                            file = self.stil_file

                    with open(file) as data:
                        self.tree = self.parser.parse(data.read())
                        if debug == True:
                            print(self.tree.pretty())

            if debug:
                print("\nEnd of syntax parsing")
                print("===========================\n")

            return self.tree

        except UnexpectedToken as e:

#            raise Exception()
#            print(f"UnexpectedToken Exception:\n {e.token.type}")

            if debug:
                print(f"UnexpectedToken Exception:\n {e}")

            line = e.line
            self.err_line = line

            col = e.column
            self.err_col = col
            
            line_with_error = utils.get_line(self.stil_file, e.line)

            self.err_msg = f"\nERROR:\tIn the line number {line}, column {col}:\n"
            self.err_msg += f"\t{line_with_error}\n"
            self.err_msg += utils.get_col_error_pos(col) + "\n"

            if str(e).startswith("Unexpected token Token('$END'"):
                s = str(e).split("\n")
                i = 0
                for s1 in s:
                    i += 1
                    if i == 3:
                        strim = s1.strip()
                        if strim == "*":
                            lines = line_with_error.strip()
                            exp_token = "';' or '}'"
                            if (lines.endswith(";")):
                                exp_token = "'}'"
                            self.err_msg += f"\tEnd of the file reached, but expected {exp_token}"
                        else:
                            if strim.startswith("* "):
                                self.err_msg += "\tUnexpected end of the file or block is not closed (including nested ones) \n"
#                                expected = strim[2:]
#                                self.err_msg += f"\t{SyntaxParserExceptions(debug).transform([expected])}\n"
                            else:
                                self.err_msg += s1 + "\n"
            else:
                self.err_msg += f"\tUnexpected token : {e.token}\n"
                token = e.token.type
                if token == "b_pattern_burst__CLOSE_PATT_OR_BURST":
                    self.err_msg += "\tDid you miss to write the mandatory PatList block ? \n"
                else:
                    self.err_msg += f"\t{SyntaxParserExceptions(debug).transform(e.expected)}\n"

            print(self.err_msg)

        except UnexpectedCharacters as e:

            if debug:
                print(f"UnexpectedCharacters Exception:\n {e}")

            line = e.line
            self.err_line = line

            col = e.column
            self.err_col = col

            line_with_error = utils.get_line(self.stil_file, line)

            self.err_msg = f"\nERROR:\tIn the line number {line}, column {col}:\n"
            self.err_msg += f"\t{line_with_error}\n"
            self.err_msg += utils.get_col_error_pos(col) + "\n"
            self.err_msg += f"\tUnexpected character : {e.char}\n"
            self.err_msg += f"\t{SyntaxParserExceptions(debug).transform(e.allowed)}\n"

            print(self.err_msg)

        return None

    def parse_semantic(
        self,
        # stil,
        debug=False,
    ):

        # if tree is None, syntax parsing is already failed
        if self.tree != None:
            if debug:
                print("\n===========================")
                print("Start of semantic analisys :\n")
            try:
                self.transform(self.tree)
                self.eof()
                self.is_parsing_done = True

                if debug:
                    print("\nEnd of semantic analisys")
                    print("===========================\n")
                                    
            except VisitError as e:

                #raise Exception()
                #if debug:
                #    raise Exception()
                                
                if isinstance(e.orig_exc, STILSemanticException):
                    line = e.orig_exc.line
                    col = e.orig_exc.col
                    err_msg = e.orig_exc.msg
                else:

                    err_msg = e.orig_exc
                    try:
                        #Lark version < 1.0.0 
                        line = e.obj.line
                        col = e.obj.column
                    except:
                        #Lark version >= 1.0.0 
                        line = e.obj.meta.line
                        col = e.obj.meta.column
                
                self.err_line = line
                self.err_col = col

                line_with_error = utils.get_line(self.stil_file, line)
                self.err_msg = f"\nERROR:\tIn the line number {line}, column {col} [{line},{col}]:\n"
                self.err_msg += f"\t{line_with_error}\n"
                self.err_msg += utils.get_col_error_pos(col) + "\n"
                self.err_msg += f"\t{err_msg}\n"

                print(self.err_msg)

        else:
            print("ERROR : syntax parsing must be performed before semantic analisys b")



# -*- coding: utf-8 -*-
import inspect

from .DomainUtils import DomainUtils

class SignalGroupsBlockParser:
    def __init__(self, debug=False):
        self.debug = debug
        # List with all parsed signal groups domains
        self.signal_groups_domains = []
        # dict key   is the signal group name with the domain -> domain::signal_group
        # dict value is a list with all signals defined for the key signal group
        self.signal_groups2signals = {}

        # Current full signal group name (domain::signal_group)
        self.curr_sig_group_name = None
        # Current domain name
        self.curr_sig_group_domain = DomainUtils.global_domain
        self.last_op = None
        # All signal groups names without domain
        # Will be used in Macro/Proc blocks validation
        self.all_signal_groups = []

        self.is_signalgroups_block_defined = False
        
        # dict key   is the signal_groups_domain::signal name
        # dict value is the default drive state of the signal if not used
        self.sgd_sig2def_state = {}


    def trace(self, func_name, t):
        head = f"{__name__}:{func_name}"

        if isinstance(t, list):
            print(f"{head} rule value {t}")
        else:
            print(f'{head} token value "{t}" at line {t.line} column {t.column}')

    def b_signal_groups__SIGNAL_GROUPS_DOMAIN_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_sig_group_domain = t.value

    def b_signal_groups__OPEN_SIGNAL_GROUPS_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        if self.is_signal_block_defined == False:
            err_msg = "Expected Signal block before any SignalGroups block"
            raise Exception(err_msg)
        self.is_signalgroups_block_defined = True

        if self.curr_sig_group_domain in self.signal_groups_domains:
            domain_name = DomainUtils.get_domain(self.curr_sig_group_domain, True)
            err_msg = f"SignalGroups block domain '{domain_name}' is already defined"
            raise Exception(err_msg)
        else:
            self.signal_groups_domains.append(self.curr_sig_group_domain)

    def b_signal_groups__SIGNAL_GROUP_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        sig_group_name = t.value
        if sig_group_name in self.sig2type:
            err_msg = f"Signal group name '{sig_group_name}' is already defined as signal in the Signals block"
            raise Exception(err_msg)
        self.curr_sig_group_name = DomainUtils.get_full_name(
            self.curr_sig_group_domain, t.value
        )

        if self.curr_sig_group_name in self.signal_groups2signals:
            err_msg = f"Signal group name '{sig_group_name}' is already defined"
            raise Exception(err_msg)
        else:
            self.signal_groups2signals[self.curr_sig_group_name] = []

        if sig_group_name not in self.all_signal_groups:
            self.all_signal_groups.append(sig_group_name)
            
            

    def b_signal_groups__SIG_ADD(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.last_op = t.value

    def b_signal_groups__SIG_SUB(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.last_op = t.value

    def b_signal_groups__SIGREF_NAME(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        sig_name = t.value
        sig_list = self.signal_groups2signals[self.curr_sig_group_name]
        sig_group = DomainUtils.get_full_name(self.curr_sig_group_domain, sig_name)

        if sig_name in self.sig2type:
            # if the sigref is a signal
            if self.last_op == None:
                sig_list.append(sig_name)
            elif self.last_op == "+":
                if sig_name in sig_list:
                    name = DomainUtils.get_name(self.curr_sig_group_name)
                    err_msg = f"In the signal group '{name}', can not add already existing signal '{sig_name}' from previously used signal groups or signal in the current signals group"
                    raise Exception(err_msg)
                else:
                    sig_list.append(sig_name)
            elif self.last_op == "-":
                if sig_name not in sig_list:
                    name = DomainUtils.get_name(self.curr_sig_group_name)
                    err_msg = f"In the signal group '{name}', can not remove not existing signal '{sig_name}' from previously used signal groups or signal in the current signals group"
                    raise Exception(err_msg)
                else:
                    sig_list.remove(sig_name)
            else:
                err_msg = "Internal error, expecting '+' or '-' operator"
                raise Exception(err_msg)

        elif sig_group in self.signal_groups2signals:
            # if the sigref is a signal group
            signals = self.signal_groups2signals[sig_group]
            if self.last_op == None:
                sig_list.extend(signals)
            elif self.last_op == "+":
                check = any(item in sig_list for item in signals)
                if check:
                    name = DomainUtils.get_name(self.curr_sig_group_name)
                    err_msg = f"Can not add signal group '{sig_name}' into signal group '{name}' because at least one signal is duplicated in both groups"
                    raise Exception(err_msg)
                else:
                    sig_list.extend(signals)
            elif self.last_op == "-":
                check = all(item in sig_list for item in signals)
                if check:
                    for s in signals:
                        sig_list.remove(s)
                else:
                    name = DomainUtils.get_name(self.curr_sig_group_name)
                    err_msg = f"Can not remove signal group '{sig_name}' from signal group '{name}' because not all signals from '{sig_name}' exists in previously used signal groups"
                    raise Exception(err_msg)
        else:
            # raise an error, because the sigref is not defined
            err_msg = f"Signal {sig_name} is not defined in the Signals block"
            raise Exception(err_msg)
        self.last_op = None

    def b_signal_groups__SIG_DEF_STATE_UP(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.sig2def_state[self.curr_sig_group_name] = "U"

    def b_signal_groups__SIG_DEF_STATE_DOWN(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.sig2def_state[self.curr_sig_group_name] = "D"

    def b_signal_groups__SIG_DEF_STATE_OFF(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)

        self.sig2def_state[self.curr_sig_group_name] = "Z"
        
    def b_signal_groups__SIGNAL_LIST_END_STMT(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        sigs = self.signal_groups2signals[self.curr_sig_group_name]
        if len(sigs) == 0:
            sig_group = DomainUtils.get_name(self.curr_sig_group_name)
            err_msg = f"Signal group {sig_group} does not have signals"
            raise Exception(err_msg)
        self.curr_sig_group_name = None

    def b_signal_groups__CLOSE_SIGNAL_ATTR_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_sig_group_name = None

    def b_signal_groups__CLOSE_SIGNAL_GROUPS_BLOCK(self, t):
        if self.debug:
            func_name = inspect.stack()[0][3]
            self.trace(func_name, t)
        self.curr_sig_group_domain = DomainUtils.global_domain
